import java.awt.*;

/**
   This class displays an animation on a component that can be added to any
   container

   @author David Winchurch.
   @version 09/03/2001
**/

public class AnimatedPanel extends Component
{
   //The graphics primatives required to double buffer the animation
   private Graphics backSurface;
   private Image backBuffer;

   //The animation frames are stored here
   private Image[] animation;

   //pointer to the array position of the current frame
   private int currentFrame;

   /**
      Constructor. Creates and instantiates all the global variables

      @param anim the frames of animation you want to be displayed stored in an Image array
   **/

   public AnimatedPanel(Image[] anim)
   {
      animation=anim;
      currentFrame=0;
   }

   /**
      Required for pack methods to work

      @returns the prefered size of the component as a Dimension
   **/

   public Dimension getPreferredSize()
   {
      return new Dimension(animation[0].getWidth(this),animation[0].getHeight(this));
   }

   /**
      Moves to the next frame in the animation and redraws the component
   **/

   public void nextFrame()
   {
      if(currentFrame==animation.length-1) currentFrame=0; else currentFrame++;
   }

   /**
      Overwritten paint method draws the current frame of the animation to the
      screen

      @param g the graphics context to use for painting.
   **/

   public void paint(Graphics g)
   {
      //creates the graphics primatives for the backbuffer provided it will be legal to create them
      if(backBuffer==null && getSize().width>0 && getSize().height>0)
      {
         backBuffer=createImage(getSize().width, getSize().height);
         backSurface=backBuffer.getGraphics();
      }

      //Clears the backbuffer
      backSurface.clearRect(0,0, getSize().width, getSize().height);

      //Draws the current frame of animation to the backbuffer
      backSurface.drawImage(animation[currentFrame], 0, 10, this);

      //Draws the backbuffer to the screen
      g.drawImage(backBuffer, 0,0, Color.gray, this);
   }
} 