import java.awt.*;
import java.io.*;
import java.net.URL;
import java.util.zip.*;

/**
   This class contains all the methods required to decode and run a gameboy rom
   on a java enabled machine. For a detailed explanation of how this class works
   please refer to section FILL IN HERE!!!!!!!!!!!!!!

   @author David Winchurch.
   @version 11/03/2001i
**/

public class GameBoyCPU implements Runnable
{
   //The virtual memory arrays that store the loaded rom and perform the functions
   //of the gameboys memory
   private char GB_MEM[]=new char[0x10000],
                VRAM[]=new char[0x4000],
                CART_RAM[],
                PAGED_RAM[]=new char[0x8000],
                CART_ROM[];

   //The virtual video subsystems
   private GameBoyVideo VIDEO;

   private char A,B,C,D,E,F,H,L, //The vitual 8 bit registers
                SWAP_TABLE[], //Array used for the SWAP instruction
                tempChar;

   private int AF,BC,DE,HL,PC,SP, //The virtual 16 bit registers
               DAA_TABLE[], //Array used for the DAA instruction
               MBC, MBC_MODE, RAM_PAGE, ROM_PAGE, TIMER, TIMER_MAX,
               //Internal variables required for certain subsystem emulation
               CPU_CYCLES, CPU_SPEED, TIMER_CYCLES, DIV, SPR_PER_LINE, opCode, opCode2, tempOpCodeRegister;

   private boolean APPLET_RUNNING, //Controls the start stop state of the thread
                   KEYS_PRESSED[], //Stores which virtual joypad buttons have been pressed
                   IME, //Interrupt master enable
                   STOP, HALT, //Controls the state of the cpu
                   //Internal variables required for certain subsystem emulation
                   HALT_NEXT, TIMER_ON, RAM_ENABLE, RUMBLE_PACK;

   //A reference to the colours used in black and white emulation mode
   private Color[][] DEFAULTS={{Color.white, Color.lightGray},{Color.gray, Color.black}},
                     BGPALETTE={{Color.white, Color.lightGray},{Color.gray, Color.black}},
                     OBJ0PALETTE={{Color.white, Color.lightGray},{Color.gray, Color.black}},
                     OBJ1PALETTE={{Color.white, Color.lightGray},{Color.gray, Color.black}};

   //The thread
   private Thread instance;

   //lookup table for all the opcode timings
   private final int[] incAllValue={4, 12,8, 8, 4, 4, 8, 4, 20,8, 8, 8, 4, 4, 8, 4,
                                    4, 12,8, 8, 4, 4, 8, 4, 12,8, 8, 8, 4, 4, 8, 4,
                                    8, 12,8, 8, 4, 4, 8, 4, 8, 8, 8, 8, 4, 4, 8, 4,
                                    8, 12,8, 8, 8, 8, 12,4, 8, 8, 8, 8, 4, 4, 8, 4,
                                    4, 4, 4, 4, 4, 4, 8, 4, 4, 4, 4, 4, 4, 4, 8, 4,
                                    4, 4, 4, 4, 4, 4, 8, 4, 4, 4, 4, 4, 4, 4, 8, 4,
                                    4, 4, 4, 4, 4, 4, 8, 4, 4, 4, 4, 4, 4, 4, 8, 4,
                                    8, 8, 8, 8, 8, 8, 4, 8, 4, 4, 4, 4, 4, 4, 8, 4,
                                    4, 4, 4, 4, 4, 4, 8, 4, 4, 4, 4, 4, 4, 4, 8, 4,
                                    4, 4, 4, 4, 4, 4, 8, 4, 4, 4, 4, 4, 4, 4, 8, 4,
                                    4, 4, 4, 4, 4, 4, 8, 4, 4, 4, 4, 4, 4, 4, 8, 4,
                                    4, 4, 4, 4, 4, 4, 8, 4, 4, 4, 4, 4, 4, 4, 8, 4,
                                    8, 12,12,16,12,16,4, 16,8, 20,12,0, 12,24,4, 16,
                                    8, 12,12,0 ,12,16,4, 16,8, 20,12,0, 12,0, 4, 16,
                                    12,12,8, 0, 0, 16,4, 16,16,4, 4, 0, 0, 0, 4, 16,
                                    12,12,8, 4, 0, 16,4, 16,12,8, 4, 4, 0, 0, 4, 16};

   //lookup table for all the CB opcode timings
   private final int[] incAllValueCB={8, 8, 8, 8, 8, 8, 16, 8, 8, 8, 8, 8, 8, 8, 16, 8,
                                      8, 8, 8, 8, 8, 8, 16, 8, 8, 8, 8, 8, 8, 8, 16, 8,
                                      8, 8, 8, 8, 8, 8, 16, 8, 8, 8, 8, 8, 8, 8, 16, 8,
                                      8, 8, 8, 8, 8, 8, 16, 8, 8, 8, 8, 8, 8, 8, 16, 8,
                                      8, 8, 8, 8, 8, 8, 16, 8, 8, 8, 8, 8, 8, 8, 16, 8,
                                      8, 8, 8, 8, 8, 8, 16, 8, 8, 8, 8, 8, 8, 8, 16, 8,
                                      8, 8, 8, 8, 8, 8, 16, 8, 8, 8, 8, 8, 8, 8, 16, 8,
                                      8, 8, 8, 8, 8, 8, 16, 8, 8, 8, 8, 8, 8, 8, 16, 8,
                                      8, 8, 8, 8, 8, 8, 16, 8, 8, 8, 8, 8, 8, 8, 16, 8,
                                      8, 8, 8, 8, 8, 8, 16, 8, 8, 8, 8, 8, 8, 8, 16, 8,
                                      8, 8, 8, 8, 8, 8, 16, 8, 8, 8, 8, 8, 8, 8, 16, 8,
                                      8, 8, 8, 8, 8, 8, 16, 8, 8, 8, 8, 8, 8, 8, 16, 8,
                                      8, 8, 8, 8, 8, 8, 16, 8, 8, 8, 8, 8, 8, 8, 16, 8,
                                      8, 8, 8, 8, 8, 8, 16, 8, 8, 8, 8, 8, 8, 8, 16, 8,
                                      8, 8, 8, 8, 8, 8, 16, 8, 8, 8, 8, 8, 8, 8, 16, 8,
                                      8, 8, 8, 8, 8, 8, 16, 8, 8, 8, 8, 8, 8, 8, 16, 8};

   //The dialog box used to display the loading rom window
   private Dialog loadingDialog;

   //Components for the loading rom window
   private AnimatedPanel loadingAnim;
   private PercentageBar romLoading, ramLoading;

   //Stores the filename of the rom
   private String romName;

   //Uncheangeable values storing the flag bits
   private final char Z_FLAG = 0x80; //1000 0000
   private final char N_FLAG = 0x40; //0100 0000
   private final char H_FLAG = 0x20; //0010 0000
   private final char C_FLAG = 0x10; //0001 0000

   //Uncheangeable values reprisenting important gameboy registers
   private final int JOYPAD                = 0xFF00;
   private final int SERIAL_DATA           = 0xFF01;
   private final int SERIAL_CTRL           = 0xFF02;
   private final int DIV_CNTR              = 0xFF04;
   private final int TIMER_COUNT           = 0xFF05;
   private final int TIMER_RELOAD          = 0xFF06;
   private final int TIMER_CRTL            = 0xFF07;
   private final int INT_FLAG              = 0xFF0F;

   private final int SND_1_ENT             = 0xFF10;
   private final int SND_1_WAV_LEN         = 0xFF11;
   private final int SND_1_ENV             = 0xFF12;
   private final int SND_1_FREQ_KICK_LOWER = 0xFF13;
   private final int SND_1_FREQ_KICK_UPPER = 0xFF14;
   private final int SND_2_WAVE_LEN        = 0xFF16;
   private final int SND_2_ENV             = 0xFF17;
   private final int SND_2_FREQ_KICK_LOWER = 0xFF18;
   private final int SND_2_FREQ_KICK_UPPER = 0xFF19;
   private final int SND_3_ON_OFF          = 0xFF1A;
   private final int SND_3_LEN             = 0xFF1B;
   private final int SND_3_VOLUME          = 0xFF1C;
   private final int SND_3_FREQ_KICK_LOWER = 0xFF1D;
   private final int SND_3_FREQ_KICK_UPPER = 0xFF1E;

   private final int SND_4_LEN             = 0xFF20;
   private final int SND_4_ENV             = 0xFF21;
   private final int SND_4_POLY_KICK_LOWER = 0xFF22;
   private final int SND_4_POLY_KICK_UPPER = 0xFF23;
   private final int SND_VOICE_INP         = 0xFF24;
   private final int SND_STEREO            = 0xFF25;
   private final int SND_STAT              = 0xFF26;

   private final int SND_BNK_10            = 0xFF30;
   private final int SND_BNK_11            = 0xFF31;
   private final int SND_BNK_12            = 0xFF32;
   private final int SND_BNK_13            = 0xFF33;
   private final int SND_BNK_14            = 0xFF34;
   private final int SND_BNK_15            = 0xFF35;
   private final int SND_BNK_16            = 0xFF36;
   private final int SND_BNK_17            = 0xFF37;
   private final int SND_BNK_20            = 0xFF38;
   private final int SND_BNK_21            = 0xFF39;
   private final int SND_BNK_22            = 0xFF3A;
   private final int SND_BNK_23            = 0xFF3B;
   private final int SND_BNK_24            = 0xFF3C;
   private final int SND_BNK_25            = 0xFF3D;
   private final int SND_BNK_26            = 0xFF3E;
   private final int SND_BNK_27            = 0xFF3F;

   private final int LCD_CTRL              = 0xFF40;
   private final int LCD_STAT              = 0xFF41;
   private final int LCD_SCROLL_Y          = 0xFF42;
   private final int LCD_SCROLL_X          = 0xFF43;
   private final int LCD_Y_LOC             = 0xFF44;
   private final int LCD_Y_COMP            = 0xFF45;
   private final int LCD_DMA               = 0xFF46;
   private final int LCD_BACK_PALETTE      = 0xFF47;
   private final int LCD_SPR0_PALETTE      = 0xFF48;
   private final int LCD_SPR1_PALETTE      = 0xFF49;
   private final int LCD_WIN_Y             = 0xFF4A;
   private final int LCD_WIN_X             = 0xFF4B;
   private final int CPU_SPEED_REG         = 0xFF4D; //GBC
   private final int VRAM_BANK             = 0xFF4F; //GBC

   private final int DMA_SRC_UPPER         = 0xFF51; //GBC
   private final int DMA_SRC_LOWER         = 0xFF52; //GBC
   private final int DMA_DST_UPPER         = 0xFF53; //GBC
   private final int DMA_DST_LOWER         = 0xFF54; //GBC
   private final int DMA_LEN_TYPE          = 0xFF55; //GBC
   private final int IR_PORT               = 0xFF56; //GBC

   private final int BGP_INDEX             = 0xFF68; //GBC
   private final int BGP_DATA              = 0xFF69; //GBC
   private final int OBP_INDEX             = 0xFF6A; //GBC
   private final int OBP_DATA              = 0xFF6B; //GBC

   private final int RAM_BANK              = 0xFF70; //GBC
   private final int INT_ENABLE            = 0xFFFF;

   //Uncheangeable values reprisenting keys
   private final int DIK_UP     = 0;
   private final int DIK_DOWN   = 1;
   private final int DIK_LEFT   = 2;
   private final int DIK_RIGHT  = 3;
   private final int DIK_A      = 4;
   private final int DIK_B      = 5;
   private final int DIK_START  = 6;
   private final int DIK_SELECT = 7;

   /**
      Constructor. Sets up the loading rom dialog box

      @param keys pointer to the keys pressed array
      @param im the images used by the animated panel on the rom loading dialog box
   **/

   public GameBoyCPU(boolean[] keys, Image[] im)
   {
      createDAATable();
      createSwapTable();

      //Creates a new instance of the virtual video subsystems
      VIDEO=new GameBoyVideo(GB_MEM, VRAM);
      KEYS_PRESSED=keys;

      //Sets up the loading rom dialog box
      loadingDialog=new Dialog(new Frame(), "Loading...", false)
      {
         public void paintAll(Graphics g)
         {
            loadingDialog.paint(loadingDialog.getGraphics());
            romLoading.paint(romLoading.getGraphics());
            ramLoading.paint(ramLoading.getGraphics());
         }
      };
      loadingDialog.setLocation(200, 200);
      loadingDialog.setLayout(new BorderLayout());
      loadingDialog.setResizable(false);

      //creates the components used in the loading dialog box
      loadingAnim=new AnimatedPanel(im);
      romLoading=new PercentageBar("Rom Loading");
      ramLoading=new PercentageBar("Ram Loading");

      //Creates and adds components to a panel
      Panel status=new Panel(new GridLayout(2,1));
      status.add(romLoading);
      status.add(ramLoading);

      //Add the components to their relevant locations
      loadingDialog.add(loadingAnim, BorderLayout.WEST);
      loadingDialog.add(status, BorderLayout.CENTER);

      //resizes the window to its optimum size
      loadingDialog.pack();
   }

   /**
      Starts the thread and causes the virtual cpu to start
   **/

   public void startThread()
   {
      if(instance==null)
      {
         APPLET_RUNNING=true;
         instance=new Thread(this);
         instance.start();
      }
   }

   /**
      Stops the thread and causes the virtual cpu to stop
   **/

   public void stopThread()
   {
      APPLET_RUNNING=false;
      instance=null;
   }

   /**
      Called when the thread starts constantly calls opCodeList1()
   **/

   public void run()
   {
      instance.setPriority(Thread.MIN_PRIORITY);
      while (APPLET_RUNNING)
         opCodeList1();
   }

   /**
      This method loads a rom into the virtual memory array it also loads any
      save data present plus any palette info if present and the rom is B&W

      @param fileName the full path and filename of a rom
      @return true should the loading of the rom be succesful
   **/

   public boolean loadRom(URL url, String fileName)
   {
      //Stores the state of the loading process
      boolean romLoaded;

      try
      {
         //Attempts to open a zip file from the internet
         ZipInputStream zipInput=new ZipInputStream(new URL(url, fileName).openStream());
         zipInput.getNextEntry(); //gets the first file in the zip folder

         //displays the loading rom screen
         loadingDialog.setVisible(true);

         ///these must be called to make the percentage bars display
         romLoading.setPercentage(0f);
         ramLoading.setBarColor(Color.gray); ramLoading.setPercentage(1f);
         loadingDialog.paintAll(loadingDialog.getGraphics());

         //integers used to keep track of loading
         int fileSize=0, ramSize=0, memLoc=0x8000, ramLoc=0;

         //gets just the filename of the rom without extensions
         for(int i=0; i<fileName.length(); i++) if(fileName.charAt(i)=='.') romName=fileName.substring(0, i);

         //temporary byte used to store the data as its loaded from the file
         byte inputStreamByte[]=new byte[1];

         //loads the first 0x8000 bytes into the virtual memory. this represents the
         //first 2 pages of the rom. the minimum in any rom
         for(int i=0; i<0x8000; i++)
         {
            //checks to see that the loader is not at the end of the file
            if(zipInput.read(inputStreamByte,0,1)==-1)
            {
               System.out.println("The rom file is smaller than it should be and may be corrupt");
               break;
            }
            else
            {
               //removes the sign for the loaded byte and copies it to its relevant location in virtual memory
               GB_MEM[i]=(char)(inputStreamByte[0]>=0? inputStreamByte[0]:(inputStreamByte[0]+256));
            }
         }

         //calculates the file size needed to store rom in
         switch(GB_MEM[0x148])
         {
            case 0x00 : {fileSize=0x8000; break;}  //256Kbit=32KByte=2 banks
            case 0x01 : {fileSize=0x10000; break;}  //512Kbit=64KByte=4 banks
            case 0x02 : {fileSize=0x20000; break;}  //1Mbit=128KByte=8 banks
            case 0x03 : {fileSize=0x40000; break;}  //2Mbit=256KByte=16 banks
            case 0x04 : {fileSize=0x80000; break;}  //4Mbit=512KByte=32 banks
            case 0x05 : {fileSize=0x100000; break;} //8Mbit=1MByte=64 banks
            case 0x06 : {fileSize=0x200000; break;} //16Mbit=2MByte=128 banks
            case 0x07 : {fileSize=0x400000; break;} //32Mbit=4MByte=256 banks
            case 0x52 : {fileSize=0x120000; break;} //9Mbit=1.1MByte=72 banks
            case 0x53 : {fileSize=0x140000; break;} //10Mbit=1.2MByte=80 banks
            case 0x54 : {fileSize=0x180000; break;} //12Mbit=1.5MByte=96 banks

            default : {System.out.println("Unknown file size"); break;}
         }

         //Creates a virtual memory array with enough space to load the whole rom
         CART_ROM=new char[fileSize];

         //Copies the pre loaded section into the new array
         for(int i=0; i<0x8000; i++)
         {
            CART_ROM[i]=GB_MEM[i];
         }

         try
         {
            //removes the sign for the loaded byte and copies it to its relevant location in virtual memory
            while (zipInput.read(inputStreamByte,0,1)!=-1)
            {
               CART_ROM[memLoc++]=(char)(inputStreamByte[0]>=0? inputStreamByte[0]:(inputStreamByte[0]+256));

               //Updates the percentage bar
               if((memLoc&0xFFF)==0x00)
               {
                  romLoading.setPercentage((float)memLoc/fileSize); loadingAnim.nextFrame();
                  loadingDialog.paintAll(loadingDialog.getGraphics());
               }
            }
         }
         catch(ArrayIndexOutOfBoundsException ex) {System.out.println("The rom file is larger than it should be and may be corrupt");}

         //Closes the file
         zipInput.close();

         //calculates the file size needed to store the ram in
         switch(GB_MEM[0x149])
         {
            case 0x00 : {ramSize=0x00; break;}
            case 0x01 : {ramSize=0x500; break;}
            case 0x02 : {ramSize=0x2000; break;}
            case 0x03 : {ramSize=0x8000; break;}
            case 0x04 : {ramSize=0x16000; break;}

            default : {System.out.println("Unknown ram size"); break;}
         }

         if(ramSize>0)
         {
            //Creates a virtual memory array with enough space to load the saved data
            CART_RAM=new char[ramSize];
            try
            {
               DataInputStream ramInput=new DataInputStream(new URL(url, romName+".sav").openStream());
               try
               {
                  ramLoading.setBarColor(new Color(0,0,128)); ramLoading.setPercentage(0f);
                  //removes the sign for the loaded byte and copies it to its relevant location in virtual memory
                  while (ramInput.read(inputStreamByte,0,1)!=-1)
                  {
                     CART_RAM[ramLoc++]=(char)(inputStreamByte[0]>=0? inputStreamByte[0]:(inputStreamByte[0]+256));

                     //Updates the percentage bar
                     if((ramLoc&0xFF)==0x00)
                     {
                        ramLoading.setPercentage((float)ramLoc/ramSize); loadingAnim.nextFrame();
                        loadingDialog.paintAll(loadingDialog.getGraphics());
                     }
                  }
               }
               catch(ArrayIndexOutOfBoundsException ex) {System.out.println("The ram file is larger than it should be and may be corrupt");}

               //Closes the file
               ramInput.close();
            }
            catch (IOException ex) {}
         }
         else CART_RAM=null;


         if(GB_MEM[0x0143]<0x80)
         {
            try
            {
               //Creates a new BurreredReader from a file on the hard disk. Throws an error if the file is not present
               DataInputStream paletteReader=new DataInputStream(new URL(url, romName+".pal").openStream());

               //Used to track the number of lines in the text file
               int paletteFileSize=0;
               //Loops until the end of the file
               while (paletteReader.readLine()!=null) paletteFileSize++;

               //Recreates the BufferedReader. Used to start the reader at the begining of the file again
               paletteReader=new DataInputStream(new URL(url, romName+".pal").openStream());

               for(int i=0; i<paletteFileSize; i++)
               {
                  String tempString=paletteReader.readLine();
                  //loops through the string checking each character at a time to see if it's an '='
                  for(int j=0; j<tempString.length(); j++)
                  {
                     //once it finds the required character these if statements find out if the subtring before
                     //the '=' is part of the stored user information
                     if(tempString.charAt(j)=='=')
                     {
                        if(tempString.substring(0, j).equals("bg0")) {BGPALETTE[0][0]=stringToColor(tempString.substring(j+1, tempString.length()));}
                        if(tempString.substring(0, j).equals("bg1")) {BGPALETTE[0][1]=stringToColor(tempString.substring(j+1, tempString.length()));}
                        if(tempString.substring(0, j).equals("bg2")) {BGPALETTE[1][0]=stringToColor(tempString.substring(j+1, tempString.length()));}
                        if(tempString.substring(0, j).equals("bg3")) {BGPALETTE[1][1]=stringToColor(tempString.substring(j+1, tempString.length()));}

                        if(tempString.substring(0, j).equals("obj00")) {OBJ0PALETTE[0][0]=stringToColor(tempString.substring(j+1, tempString.length()));}
                        if(tempString.substring(0, j).equals("obj01")) {OBJ0PALETTE[0][1]=stringToColor(tempString.substring(j+1, tempString.length()));}
                        if(tempString.substring(0, j).equals("obj02")) {OBJ0PALETTE[1][0]=stringToColor(tempString.substring(j+1, tempString.length()));}
                        if(tempString.substring(0, j).equals("obj03")) {OBJ0PALETTE[1][1]=stringToColor(tempString.substring(j+1, tempString.length()));}

                        if(tempString.substring(0, j).equals("obj10")) {OBJ1PALETTE[0][0]=stringToColor(tempString.substring(j+1, tempString.length()));}
                        if(tempString.substring(0, j).equals("obj11")) {OBJ1PALETTE[0][1]=stringToColor(tempString.substring(j+1, tempString.length()));}
                        if(tempString.substring(0, j).equals("obj12")) {OBJ1PALETTE[1][0]=stringToColor(tempString.substring(j+1, tempString.length()));}
                        if(tempString.substring(0, j).equals("obj13")) {OBJ1PALETTE[1][1]=stringToColor(tempString.substring(j+1, tempString.length()));}
                     }
                  }
               } 
               //Closes the file
               paletteReader.close();
            }
            catch(IOException ex)
            {
               //Any errors cause default palette data to be loaded
               BGPALETTE[0][0]=DEFAULTS[0][0];
               BGPALETTE[0][1]=DEFAULTS[0][1];
               BGPALETTE[1][0]=DEFAULTS[1][0];
               BGPALETTE[1][1]=DEFAULTS[1][1];
               OBJ0PALETTE[0][0]=DEFAULTS[0][0];
               OBJ0PALETTE[0][1]=DEFAULTS[0][1];
               OBJ0PALETTE[1][0]=DEFAULTS[1][0];
               OBJ0PALETTE[1][1]=DEFAULTS[1][1]; 
               OBJ1PALETTE[0][0]=DEFAULTS[0][0];
               OBJ1PALETTE[0][1]=DEFAULTS[0][1];
               OBJ1PALETTE[1][0]=DEFAULTS[1][0];
               OBJ1PALETTE[1][1]=DEFAULTS[1][1];
            }

            //forces the rom to recalculate the color values its using
            setBackPalette();
            setForePalette0();
            setForePalette1();
         }

         //hides the rom loading window
         loadingDialog.setVisible(false);

         //If the method has got this far the rom has been loaded sucessfully
         romLoaded=true;
      }
      catch (IOException ex) {romLoaded=false;}

      //return the succesfulness of the rom loading procedure
      return romLoaded;
   }

   /**
      Converts a String of the style redvalue,greenvalue,bluevalue into a
      colour

      @param colString a string representing to components of a color
      @return the color reprisented by the string
   **/

   public Color stringToColor(String colString)
   {
      String rgbTemp="";

      //array to store the colour component values in
      int[] rgb=new int[3];
      int rgbPointer=0;

      for(int k=0; k<colString.length(); k++)
      {
         if(colString.charAt(k)==',')
         {
            rgb[rgbPointer]=Integer.parseInt(rgbTemp);
            rgbPointer++;
            rgbTemp="";
         }
         else
            rgbTemp+=colString.charAt(k);
      }
      rgb[rgbPointer]=Integer.parseInt(rgbTemp);

      //creates the final color
      return new Color(rgb[0], rgb[1], rgb[2]);
   }

   /**
      Resets all the important virtual memory locations and subsystems back to
      the values they would have in the original gameboy
   **/

   public void fullReset()
   {
      //The 16 registry default values
      AF=0x11B0; //GBC
      BC=0x0013;
      DE=0x00D8;
      HL=0x014D;
      SP=0xFFFE;
      PC=0x0100;

      //Ensures that the 8 bit values contain the correct values
      A=getUpper(AF);      F=getLower(AF);
      B=getUpper(BC);      C=getLower(BC);
      D=getUpper(DE);      E=getLower(DE);
      H=getUpper(HL);      L=getLower(HL);

      //Resets the non recreated arrays to zero values
      for(int i=0; i<(GB_MEM.length-0x8000); i++) GB_MEM[0x8000+i]=0;
      for(int i=0; i<VRAM.length; i++) VRAM[i]=0;
      for(int i=0; i<PAGED_RAM.length; i++) PAGED_RAM[i]=0;

      //Sets the internal variables to default values
      MBC=0;
      MBC_MODE=0;
      RAM_PAGE=0;
      ROM_PAGE=0;
      TIMER=0;
      TIMER_MAX=0;
      CPU_CYCLES=0;
      CPU_SPEED=0;
      TIMER_CYCLES=0;
      DIV=0;
      SPR_PER_LINE=0;
      IME=false;
      STOP=false;
      HALT=false;
      HALT_NEXT=false;
      TIMER_ON=false;
      RAM_ENABLE=false;
      RUMBLE_PACK=false;

      //Certain virtual memory locations must have specific values
      GB_MEM[JOYPAD]=0xCF;
      GB_MEM[SERIAL_DATA]=0x00;
      GB_MEM[SERIAL_CTRL]=0x7E;
      GB_MEM[0xFF03]=0xFF;
      GB_MEM[DIV_CNTR]=0xAF;
      GB_MEM[TIMER_COUNT]=0x00;
      GB_MEM[TIMER_RELOAD]=0x00;
      GB_MEM[TIMER_CRTL]=0xF8;
      GB_MEM[0xFF08]=0xFF;
      GB_MEM[0xFF09]=0xFF;
      GB_MEM[0xFF0A]=0xFF;
      GB_MEM[0xFF0B]=0xFF;
      GB_MEM[0xFF0C]=0xFF;
      GB_MEM[0xFF0D]=0xFF;
      GB_MEM[0xFF0E]=0xFF;
      GB_MEM[INT_FLAG]=0x00;

      GB_MEM[SND_1_ENT]=0x80;
      GB_MEM[SND_1_WAV_LEN]=0xBF;
      GB_MEM[SND_1_ENV]=0xF3;
      GB_MEM[SND_1_FREQ_KICK_LOWER]=0xFF;
      GB_MEM[SND_1_FREQ_KICK_UPPER]=0xBF;
      GB_MEM[0xFF15]=0xFF;
      GB_MEM[SND_2_WAVE_LEN]=0x3F;
      GB_MEM[SND_2_ENV]=0x00;
      GB_MEM[SND_2_FREQ_KICK_LOWER]=0xFF;
      GB_MEM[SND_2_FREQ_KICK_UPPER]=0xBF;
      GB_MEM[SND_3_ON_OFF]=0x7F;
      GB_MEM[SND_3_LEN]=0xFF;
      GB_MEM[SND_3_VOLUME]=0x9F;
      GB_MEM[SND_3_FREQ_KICK_LOWER]=0xFF;
      GB_MEM[SND_3_FREQ_KICK_UPPER]=0xBF;
      GB_MEM[0xFF1E]=0xFF;
      GB_MEM[0xFF1F]=0xFF;

      GB_MEM[SND_4_LEN]=0xFF;
      GB_MEM[SND_4_ENV]=0x00;
      GB_MEM[SND_4_POLY_KICK_LOWER]=0x00;
      GB_MEM[SND_4_POLY_KICK_UPPER]=0xBF;
      GB_MEM[SND_VOICE_INP]=0x77;
      GB_MEM[SND_STEREO]=0xF3;
      GB_MEM[SND_STAT]=0xF1;

      GB_MEM[SND_BNK_10]=0x06;
      GB_MEM[SND_BNK_11]=0xFE;
      GB_MEM[SND_BNK_12]=0x0E;
      GB_MEM[SND_BNK_13]=0x7F;
      GB_MEM[SND_BNK_14]=0x00;
      GB_MEM[SND_BNK_15]=0xFF;
      GB_MEM[SND_BNK_16]=0x58;
      GB_MEM[SND_BNK_17]=0xDF;
      GB_MEM[SND_BNK_20]=0x00;
      GB_MEM[SND_BNK_21]=0xEC;
      GB_MEM[SND_BNK_22]=0x00;
      GB_MEM[SND_BNK_23]=0xBF;
      GB_MEM[SND_BNK_24]=0x0C;
      GB_MEM[SND_BNK_25]=0xED;
      GB_MEM[SND_BNK_26]=0x03;
      GB_MEM[SND_BNK_27]=0xF7;

      GB_MEM[LCD_CTRL]=0x91;
      GB_MEM[LCD_STAT]=0x85;
      GB_MEM[LCD_SCROLL_Y]=0x00;
      GB_MEM[LCD_SCROLL_X]=0x00;
      GB_MEM[LCD_Y_LOC]=0x00;
      GB_MEM[LCD_Y_COMP]=0x00;
      GB_MEM[LCD_DMA]=0x00;
      GB_MEM[LCD_BACK_PALETTE]=0xFC;
      GB_MEM[LCD_SPR0_PALETTE]=0xFF;
      GB_MEM[LCD_SPR1_PALETTE]=0xFF;
      GB_MEM[LCD_WIN_Y]=0x00;
      GB_MEM[LCD_WIN_X]=0x00;
      GB_MEM[0xFF4C]=0xFF;
      GB_MEM[CPU_SPEED_REG]=0x7E;
      GB_MEM[0xFF4E]=0xFF;
      GB_MEM[VRAM_BANK]=0xFE;

      GB_MEM[0xFF50]=0xFF;
      GB_MEM[DMA_SRC_UPPER]=0x00;
      GB_MEM[DMA_SRC_LOWER]=0x00;
      GB_MEM[DMA_DST_UPPER]=0x00;
      GB_MEM[DMA_DST_LOWER]=0x00;
      GB_MEM[DMA_LEN_TYPE]=0xFF;
      GB_MEM[IR_PORT]=0x00;

      GB_MEM[BGP_INDEX]=0xC0;
      GB_MEM[BGP_DATA]=0x00;
      GB_MEM[OBP_INDEX]=0xC1;
      GB_MEM[OBP_DATA]=0x00;

      GB_MEM[RAM_BANK]=0xF8;

      GB_MEM[INT_ENABLE]=0x00;

      //Works out the configuration of the roms mbc
      switch(GB_MEM[0x0147])
      {
         case 0x00 : {MBC=0; break;}//return "ROM ONLY";
     		 case 0x01 : {MBC=1; break;}//return "ROM+MBC1";
      	 case 0x02 : {MBC=1; break;}//return "ROM+MBC1+CART_RAM";
  		   case 0x03 : {MBC=1; break;}//return "ROM+MBC1+CART_RAM+BATTERY";
  		   case 0x05 : {MBC=2; break;}//return "ROM+MBC2";
  		   case 0x06 : {MBC=2; break;}//return "ROM+MBC2+BATTERY";
  		   case 0x08 : {MBC=0; break;}//return "ROM+CART_RAM";
  		   case 0x09 : {MBC=0; break;}//return "ROM+CART_RAM+BATTERY";
  		   case 0x0B : {MBC=101; break;}//return "ROM+MMM01";
  		   case 0x0C : {MBC=101; break;}//return "ROM+MMM01+SRAM";
  		   case 0x0D : {MBC=101; break;}//return "ROM+MMM01+SRAM+BATTERY";
  		   case 0x0F : {MBC=3; break;}//return "ROM+MBC3+TIMER+BATTERY";
  		   case 0x10 : {MBC=3; break;}//return "ROM+MBC3+TIMER+CART_RAM+BATTERY";
  		   case 0x11 : {MBC=3; break;}//return "ROM+MBC3";
  		   case 0x12 : {MBC=3; break;}//return "ROM+MBC3+CART_RAM";
  		   case 0x13 : {MBC=3; break;}//return "ROM+MBC3+CART_RAM+BATTERY";
  		   case 0x15 : {MBC=4; break;}//return "ROM+MBC4";
  		   case 0x16 : {MBC=4; break;}//return "ROM+MBC4+CART_RAM";
  		   case 0x17 : {MBC=4; break;}//return "ROM+MBC4+CART_RAM+BATTERY";
  		   case 0x19 : {MBC=5; break;}//return "ROM+MBC5";
  		   case 0x1A : {MBC=5; break;}//return "ROM+MBC5+CART_RAM";
  		   case 0x1B : {MBC=5; break;}//return "ROM+MBC5+CART_RAM+BATTERY";
  		   case 0x1C : {MBC=5; RUMBLE_PACK=true; break;}//return "ROM+MBC5+RUMBLE";
  		   case 0x1D : {MBC=5; RUMBLE_PACK=true; break;}//return "ROM+MBC5+RUMBLE+SRAM";
  		   case 0x1E : {MBC=5; RUMBLE_PACK=true; break;}//return "ROM+MBC5+RUMBLE+SRAM+BATTERY";
  		   case 0x1F : {/*FIND OUT*/ MBC=0; break;}//return "POCKET CAMERA";
  		   case 0xFD : {/*FIND OUT*/ MBC=0; break;}//return "BANDAI TAMA5";
  		   case 0xFE : {MBC=3; break;}//return "HUDSON HuC-3";
  		   case 0xFF : {MBC=1; break;}//return "HUDSON HuC-1";
      }

      //Defaults all the palette colors
      for(int i=0; i<4; i++)
         for(int j=0; j<8; j++)
         {
            VIDEO.setBackColour(i, j, Color.white);
            VIDEO.setForeColour(i, j, Color.white);
         }

      //Sets video and timer subsystems to defaults
      setLCDControl();
      setTimer();
   }

   /**
      Gets the virtual video subsystems instance

      @return the virtual video subsystems instance
   **/

   public GameBoyVideo getVideo()
   {
      return VIDEO;
   }

   /**
      Gets the background pallete array

      @return a 2d array containing the 4 colours used in the background palette
   **/

   public Color[][] getBGPalette()
   {
      return BGPALETTE;
   }

   /**
      Gets the foreground0 pallete array

      @return a 2d array containing the 4 colours used in the foreground0 palette
   **/

   public Color[][] getOBJ0Palette()
   {
      return OBJ0PALETTE;
   }

   /**
      Gets the foreground1 pallete array

      @return a 2d array containing the 4 colours used in the foreground1 palette
   **/

   public Color[][] getOBJ1Palette()
   {
      return OBJ1PALETTE;
   }

   /**
      Sets the background palette to the values specified by location 0xFF47 in
      GB_MEM
   **/

   public void setBackPalette()
   {
      for(int i=0; i<4; i++)
         VIDEO.setBackColour(i, 0, BGPALETTE[BIT(GB_MEM[LCD_BACK_PALETTE],(i<<1)+1)][BIT(GB_MEM[LCD_BACK_PALETTE],i<<1)]);
   }

   /**
      Sets the foreground palette 0 to the values specified by location 0xFF48 in
      GB_MEM
   **/

   public void setForePalette0()
   {
      for(int i=0; i<4; i++)
         VIDEO.setForeColour(i, 0, OBJ0PALETTE[BIT(GB_MEM[LCD_SPR0_PALETTE],(i<<1)+1)][BIT(GB_MEM[LCD_SPR0_PALETTE],i<<1)]);
   }

   /**
      Sets the foreground palette 1 to the values specified by location 0xFF49 in
      GB_MEM
   **/

   public void setForePalette1()
   {
      for(int i=0; i<4; i++)
         VIDEO.setForeColour(i, 1, OBJ1PALETTE[BIT(GB_MEM[LCD_SPR1_PALETTE],(i<<1)+1)][BIT(GB_MEM[LCD_SPR1_PALETTE],i<<1)]);
   }

   /**
      Calculates the pointers used in the video subsystem 
   **/

   private void setLCDControl()
   {
      VIDEO.setWinPointers((GB_MEM[LCD_CTRL]&0x40)>0? 0x1C00:0x1800);
      VIDEO.setTileArea((GB_MEM[LCD_CTRL]&0x10)>0? 0:0x800);
      VIDEO.setBGPointers((GB_MEM[LCD_CTRL]&0x08)>0? 0x1C00:0x1800);
   }

   /**
      Calculates the values used in the timer subsystem
   **/

   private void setTimer()
   {
      TIMER_ON=((GB_MEM[TIMER_CRTL]&0x04)>0? true:false);
      TIMER_MAX=((GB_MEM[TIMER_CRTL]&0x02)>0? ((GB_MEM[TIMER_CRTL]&0x01)>0? 256:64):((GB_MEM[TIMER_CRTL]&0x01)>0? 1024:16));
   }

   /**
      Sets certain locations in the virtual memory array specified by the virtual
      joypad input
   **/

   private void handleGBInput()
   {
      GB_MEM[JOYPAD]&=0xF0;

      if (GB_MEM[JOYPAD]==0x30)  GB_MEM[JOYPAD]=0x3F;
   	  if (GB_MEM[JOYPAD]==0x20) // LETS CHECK DOWN,UP,LEFT,RIGHT 001'0'nnnn
   	  {
         if (!KEYS_PRESSED[DIK_DOWN])  GB_MEM[JOYPAD]|=0x08;
   		   if (!KEYS_PRESSED[DIK_UP])    GB_MEM[JOYPAD]|=0x04;
  		   if (!KEYS_PRESSED[DIK_LEFT])  GB_MEM[JOYPAD]|=0x02;
   		   if (!KEYS_PRESSED[DIK_RIGHT]) GB_MEM[JOYPAD]|=0x01;
      }
   	  else if (GB_MEM[JOYPAD]==0x10) //LETS CHECK START,SELECT,B,A 00'0'1nnnn
   	  {
   		   if (!KEYS_PRESSED[DIK_START]) GB_MEM[JOYPAD]|=0x08;
   		   if (!KEYS_PRESSED[DIK_SELECT])GB_MEM[JOYPAD]|=0x04;
   		   if (!KEYS_PRESSED[DIK_B])     GB_MEM[JOYPAD]|=0x02;
   		   if (!KEYS_PRESSED[DIK_A])     GB_MEM[JOYPAD]|=0x01;
   	  }
   }

   /**
      Gets the upper 8 bits of a 16 bit value

      @param number a 16 bit number
      @return the upper 8 bits
   **/

   private char getUpper(int number)
   {
      return (char)((number>>8)&0xFF);
   }

   /**
      Gets the lower 8 bits of a 16 bit value

      @param number a 16 bit number
      @return the lower 8 bits
   **/

   private char getLower(int number)
   {
      return (char)(number&0xFF);
   }

   /**
      Combines two 8 bit numbers to create a 16 bit number

      @param upper the upper 8 bits of the final number
      @param lower the lower 8 bits of the final number
      @return the resulting 16 bit integer
   **/

   private int combine(char upper, char lower)
   {
      return (upper<<8)|lower;
   }

   /**
      Gets a specified location in the virtual memory any illegal access cause
      the cpu to stop

      @param loc the requested value
      @return the relevant value
   **/

   public char getMem(int loc)
   {
      loc&=0xFFFF; //ensures the value is 16 bit
      if ((MBC>0)&&(loc>=0x4000)&&(loc<0x8000))
      {
         //Return a value from the paged rom
         try
         {
            return CART_ROM[loc-0x4000+(ROM_PAGE<<14)];
         }
         catch(ArrayIndexOutOfBoundsException ex)
         {
            GB_MEM[CPU_SPEED_REG]=0x00;
            STOP=true;
            System.out.println("Attempted array out of bounds access in CART_ROM at rom page: " + Integer.toHexString(ROM_PAGE));
            return 0x00;
         }
      }
      else if ((loc>=0x8000)&&(loc<0xA000))
      {
         //Return a value from the vram
         try
         {
            return VRAM[(loc-0x8000)+((GB_MEM[VRAM_BANK]&0x01)<<13)];
         }
         catch(ArrayIndexOutOfBoundsException ex)
         {
            GB_MEM[CPU_SPEED_REG]=0x00;
            STOP=true;
            System.out.println("Attempted array out of bounds access in VRAM at vram page: " + Integer.toString(GB_MEM[VRAM_BANK]&0x01));
            return 0x00;
         }
      }
      else if ((MBC>0)&&(loc>=0xA000)&&(loc<0xC000)&&RAM_ENABLE)
      {
         //return a value from the cart ram
         try
         {
            return CART_RAM[(loc-0xA000)+(RAM_PAGE<<13)];
         }
         catch(ArrayIndexOutOfBoundsException ex)
         {
            GB_MEM[CPU_SPEED_REG]=0x00;
            STOP=true;
            System.out.println("Attempted array out of bounds access in CART_RAM at ram page: " + Integer.toHexString(RAM_PAGE));
            return 0x00;
         }
      }
      else if ((loc>=0xD000)&&(loc<0xE000))
      {
         //return a value from the paged gameboy ram (GAMEBOY COLOR)
         try
         {
            return PAGED_RAM[(loc-0xD000)+((GB_MEM[RAM_BANK]&0x07)<<12)];
         }
         catch(ArrayIndexOutOfBoundsException ex)
         {
            GB_MEM[CPU_SPEED_REG]=0x00;
            STOP=true;
            System.out.println("Attempted array out of bounds access in PAGED_RAM at ram bank: " + Integer.toHexString(RAM_BANK));
            return 0x00;
         }
      }
      else
      {
         //return a value from the gameboys memory
         try
         {
            return GB_MEM[loc];
         }
         catch(ArrayIndexOutOfBoundsException ex)
         {
            GB_MEM[CPU_SPEED_REG]=0x00;
            STOP=true;
            System.out.println("Attempted array out of bounds access in GB_MEM at location: " + Integer.toHexString(loc));
            return 0x00;
         }
      }
   }

   /**
      Gets a 16 bit integer from the virtual memory by combining two locations

      @return 16 bit integer from memory
   **/

   private int getMemNN()
   {
      PC+=2;
      return (getMem(PC-1)<<8)|getMem(PC-2);
   }

   /**
      Gets a signed integer from memory (normal memory isnt signed)

      @return a signed value between -128 to 127
   **/

   private int getSignedMemN()
   {
      tempChar=getMem(PC++);

      return(tempChar>127? tempChar-256:tempChar);
   }

   /**
      Sets a location in the virtual memory

      @param loc the location in memory to be altered
      @param mem the value to be set
   **/

   private void setMem(int loc, char mem)
   {
      if(loc<0x8000)
      {
         if (MBC==1) //MBC1 code
         {
            if(loc<0x2000) RAM_ENABLE=(mem&0x0F)==0x0A? true:false; //Enable ram access
            else if (loc>=0x2000 && loc<0x4000) ROM_PAGE=mem&0x1F; //Select rom page
            else if (MBC_MODE==0x01 && loc>=0x4000 && loc<0x6000) RAM_PAGE=mem&0x03; //select ram page
            else if (MBC_MODE==0x00 && loc>=0x4000 && loc<0x6000) ROM_PAGE=(ROM_PAGE&0x07)|((mem&0x03)<<3); //set the 3rd and 4th bit of the rom page selection
            else if (loc>=0x6000 && loc<0x8000) MBC_MODE=mem&0x01; //change mbc1 mode
         }
         else if (MBC==2) //MBC2 code
         {
            if (loc>=0x2000 && loc<0x4000) ROM_PAGE=mem&0x0F; //Select rom page
         }
         else if (MBC==3) //MBC3 code
         {
            if(loc<0x2000) RAM_ENABLE=(mem&0x0F)==0x0A? true:false; //Enable ram access
            else if (loc>=0x2000 && loc<0x4000)  ROM_PAGE=mem&0x7F; //Select rom page
            else if (loc>=0x4000 && loc<0x6000)  RAM_PAGE=(mem&0x03); //Select ram page
         }
         else if (MBC==5) //MBC5 code
         {
            if(loc<0x2000) RAM_ENABLE=(mem&0x0F)==0x0A? true:false; //Enable ram access
            else if (loc>=0x2000 && loc<0x3000) ROM_PAGE=(ROM_PAGE&0x100)|(mem&0xFF); //Select rom page
            else if (loc>=0x3000 && loc<0x4000) ROM_PAGE|=((mem&0x01)<<8); //set the 8th bit of the rom page selection
            else if (loc>=0x4000 && loc<0x6000)  RAM_PAGE=RUMBLE_PACK? mem&0x03:mem&0x0F; //select ram page
         }
      } 
      else if(loc==LCD_Y_LOC) {GB_MEM[LCD_Y_LOC]=0x00;} //Writes to LCD_Y_LOC reset the current horizontal scanline to 0
      else if(loc==DMA_LEN_TYPE && GB_MEM[0x0143]>=0x80) //GDMA and HDMA
      {
         GB_MEM[DMA_LEN_TYPE]=mem;

         //Copies a chunk of data into vram
         for(int i=0; i<((mem&0x7F)+1)*16; i++)
            setMem(((GB_MEM[DMA_DST_UPPER]<<8)|GB_MEM[DMA_DST_LOWER])+i, getMem(((GB_MEM[DMA_SRC_UPPER]<<8)|GB_MEM[DMA_SRC_LOWER])+i));

         GB_MEM[DMA_LEN_TYPE]&=0x80;
      }
      else if(loc==DIV_CNTR) {GB_MEM[DIV_CNTR]=0x00;} //writes to the div registed reset it
      else if(loc==LCD_CTRL) {GB_MEM[LCD_CTRL]=mem; setLCDControl();} //Writes to this location alters the lcdc properties
      else if(loc==LCD_DMA) //DMA code
      {
         GB_MEM[LCD_DMA]=mem;

         //Copies a chunk of data into vram
         for (int i=0; i<0xA0; i++)
            GB_MEM[0xFE00+i]=getMem((GB_MEM[LCD_DMA]<<8)+i);
      }
      else if(loc==JOYPAD) {GB_MEM[JOYPAD]=mem; handleGBInput();} //Writes to this location cause the gameboy to handle joypad input
      else if(loc==TIMER_CRTL) {GB_MEM[TIMER_CRTL]=mem; setTimer();} //Writes to this location alters the timer properties
      else if(loc==LCD_BACK_PALETTE && GB_MEM[0x0143]<0x80) {GB_MEM[LCD_BACK_PALETTE]=mem; setBackPalette();} //Writes to this location cause the gameboy to reconfigure the bg palette
      else if(loc==LCD_SPR0_PALETTE && GB_MEM[0x0143]<0x80) {GB_MEM[LCD_SPR0_PALETTE]=mem; setForePalette0();} //Writes to this location cause the gameboy to reconfigure the spr0 palettes
      else if(loc==LCD_SPR1_PALETTE && GB_MEM[0x0143]<0x80) {GB_MEM[LCD_SPR1_PALETTE]=mem; setForePalette1();} //Writes to this location cause the gameboy to reconfigure the spr1 palettes
      else if(loc==BGP_DATA && GB_MEM[0x0143]>=0x80) //Writes to this location cause the gameboy to set a color in one of the gameboy color BG palettes
      {
         GB_MEM[BGP_DATA]=mem;
         int pal=(GB_MEM[BGP_INDEX]&0x38)>>3;
         int colIndex=(GB_MEM[BGP_INDEX]&0x06)>>1;
         VIDEO.setBackColour(colIndex, pal, getGBColour((GB_MEM[BGP_INDEX]&0x01)>0? rgb2GBLower(VIDEO.getBackColour(colIndex, pal))|(mem<<((GB_MEM[BGP_INDEX]&0x01)*8)) : rgb2GBUpper(VIDEO.getBackColour(colIndex, pal))|mem));

         //Increments the color index if necessary
         if((GB_MEM[BGP_INDEX]&0x80)>0)
            if((GB_MEM[BGP_INDEX]&0x3F)!=0x3F)
               GB_MEM[BGP_INDEX]++;
            else GB_MEM[BGP_INDEX]&=0xC0;
      }
      else if(loc==OBP_DATA && GB_MEM[0x0143]>=0x80) //Writes to this location cause the gameboy to set a color in one of the gameboy color BG palettes
      {
         GB_MEM[OBP_DATA]=mem;
         int pal=(GB_MEM[OBP_INDEX]&0x38)>>3;
         int colIndex=(GB_MEM[OBP_INDEX]&0x06)>>1;
         VIDEO.setForeColour(colIndex, pal, getGBColour((GB_MEM[OBP_INDEX]&0x01)>0? rgb2GBLower(VIDEO.getForeColour(colIndex, pal))|(mem<<((GB_MEM[OBP_INDEX]&0x01)*8)) : rgb2GBUpper(VIDEO.getForeColour(colIndex, pal))|mem));

         //Increments the color index if necessary
         if((GB_MEM[OBP_INDEX]&0x80)>0)
            if((GB_MEM[OBP_INDEX]&0x3F)!=0x3F)
               GB_MEM[OBP_INDEX]++;
            else GB_MEM[OBP_INDEX]&=0xC0;
      }
      else if((loc>=0x8000)&&(loc<0xA000)) {VRAM[(loc-0x8000)+(0x2000*(GB_MEM[VRAM_BANK]&0x01))]=mem;} //writes to vram
      else if((loc>=0xA000)&&(loc<0xC000)&&RAM_ENABLE)
      {
         try
         {
            CART_RAM[(loc-0xA000)+(RAM_PAGE<<13)]=mem; //writes to cart ram
         }
         catch(ArrayIndexOutOfBoundsException ex)
         {
            System.out.println("Attempted array out of bounds write in CART_RAM at location: " + Integer.toHexString((loc-0xA000)+(RAM_PAGE<<13)));
         }
      }
      else if((loc>=0xC000)&&(loc<0xD000)) {GB_MEM[loc]=mem; GB_MEM[loc+0x2000]=mem;}
      else if((loc>=0xD000)&&(loc<0xE000)) {PAGED_RAM[(loc-0xD000)+(0x1000*(GB_MEM[RAM_BANK]&0x07))]=mem; if(loc<0xDE00) GB_MEM[loc+0x2000]=mem;}
      else if((loc>=0xE000)&&(loc<0xF000)) {GB_MEM[loc]=mem; GB_MEM[loc-0x2000]=mem;}
      else if((loc>=0xF000)&&(loc<0xFE00)) {GB_MEM[loc]=mem; PAGED_RAM[(loc-0xF000)+(0x1000*(GB_MEM[RAM_BANK]&0x07))]=mem;}
      else
      {
         try
         {
            GB_MEM[loc]=mem;
         }
         catch(ArrayIndexOutOfBoundsException ex)
         {
            System.out.println("Attempted array out of bounds write in GB_MEM at location: " + Integer.toHexString(loc));
         }
      }
   }

   /**
      Converts a 24 bit java color into a 15 bit gameboy color

      @param c the color for conversion
      @return the converted color
   **/

   private Color getGBColour(int c)
   {
      return new Color( (c&0x1F)*8, ((c>>5)&0x1F)*8, ((c>>10)&0x1F)*8 );
   }

   /**
      Converts a 24 bit java color into the upper 8 bits of a 15 bit gameboy color

      @param c the color for conversion
      @return the converted color
   **/

   private int rgb2GBUpper(Color c)
   {
      return (((c.getBlue()/8)<<10)+((c.getGreen()/8)<<5))&0x7F00;
   }

   /**
      Converts a 24 bit java color into the lower 8 bits of a 15 bit gameboy color

      @param c the color for conversion
      @return the converted color
   **/

   private int rgb2GBLower(Color c)
   {
      return (((c.getGreen()/8)<<5)+(c.getRed()/8))&0xFF;
   }

   /**
      The add with carry op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy

      @param src the source register
   **/

   private void ADC(char src)
   {
      tempOpCodeRegister=A+src+((F&C_FLAG)>0? 1:0);

      F=(char)(((tempOpCodeRegister&0xFF)==0? Z_FLAG:0) | (((A&0x0F)+(src&0x0F)+((F&C_FLAG)>0? 1:0))>0x0F? H_FLAG:0) | (tempOpCodeRegister>0xFF? C_FLAG:0));

      //result of addition ends up in the accumulator register
      A=getLower(tempOpCodeRegister);
      AF=combine(A,F);
   }

   /**
      The add op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy

      @param src the source register
   **/

   private void ADD(char src)
   {
      tempOpCodeRegister=A+src;

      F=(char)(((tempOpCodeRegister&0xFF)==0? Z_FLAG:0) | (((A&0x0F)+(src&0x0F))>0x0F? H_FLAG:0) | (tempOpCodeRegister>0xFF? C_FLAG:0));

      //result of addition ends up in the accumulator register
      A=getLower(tempOpCodeRegister);
      AF=combine(A,F);
   }

   /**
      The 16 bit add op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy

      @param src the source register
      @return the result of the addition
   **/

   private int ADD16(int dest, int src)
   {
      tempOpCodeRegister=(dest+src);

      F=(char)(((tempOpCodeRegister&0xFFFF)==0? Z_FLAG:0) | (((A&0x0FFF)+(src&0x0FFF))>0x0FFF? H_FLAG:0) | (tempOpCodeRegister>0xFFFF? C_FLAG:0));

      AF=combine(A,F);
      return tempOpCodeRegister&0xFFFF;
   }

   /**
      Adds a value to the stack pointer. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy

      @param src the source register
      @return the result of the addition
   **/

   private int ADDSP(int src)
   {
      tempOpCodeRegister=(SP+src);

      F=(char)(tempOpCodeRegister>0xFFFF? C_FLAG:0);

      AF=combine(A,F);
      return tempOpCodeRegister&0xFFFF;
   }

   /**
      The and op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy

      @param src the source register
   **/

   private void AND(char src)
   {
      //result of and ends up in the accumulator register
      A&=src;
      A&=0xFF;

      F=(char)(((A==0x00)? Z_FLAG:0)|H_FLAG);

      AF=combine(A,F);
   }

   /**
      Checks if a specified bit is set in a value. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy

      @param bit the tested bit
      @paran dest the tested value
   **/

   private void BIT(int bit, char dest)
   {
      if((dest&(0x01<<bit))>0)
         F&=(~Z_FLAG);
      else
         F|=Z_FLAG;

      F&=(~N_FLAG);
      F|=H_FLAG;

      AF=combine(A,F);
   }

   /**
      Checks if a specified bit is set in a value. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy

      @param num the tested value
      @paran bit the tested bit
   **/

   private int BIT(int num, int bit)
   {
      return ((num&(0x01<<bit))>0? 1:0);
   }

   /**
      The Call op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy
   **/

   private void CALL()
   {
      PUSH(PC+2);
      PC=getMemNN();
   }

   /**
      The CCF op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy
   **/

   private void CCF()
   {
      F&=(~N_FLAG);
      F&=(~H_FLAG);
      F^=C_FLAG;

      AF=combine(A,F);
   }

   /**
      The Compare op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy

      @param src the source register
   **/

   private void CP(char src)
   {
      F=(char)((A==src? Z_FLAG:0) | N_FLAG | ((A&0x0F)<(src&0x0F)? H_FLAG:0) | (A<src? C_FLAG:0));

      AF=combine(A,F);
   }

   /**
      The CPL op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy
   **/

   private void CPL()
   {
      A^=0xFF;
      A&=0xFF;
      F|=N_FLAG;
      F|=H_FLAG;

      AF=combine(A,F);
   }

   /**
      The DAA op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy
   **/

   private void DAA()
   {
      tempOpCodeRegister=A;
      if ((F&C_FLAG)>0) tempOpCodeRegister|=256;
      if ((F&H_FLAG)>0) tempOpCodeRegister|=512;
      if ((F&N_FLAG)>0) tempOpCodeRegister|=1024;
      A=getUpper(DAA_TABLE[tempOpCodeRegister]);

      if((DAA_TABLE[tempOpCodeRegister]&0x40)>0) F|=Z_FLAG; else F&=(~Z_FLAG);
      F&=(~H_FLAG);
      if((DAA_TABLE[tempOpCodeRegister]&0x01)>0) F|=C_FLAG; else F&=(~C_FLAG);

      AF=combine(A,F);
   }

   /**
      The Decrease op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy

      @param src the source register
   **/

   private char DEC(char dest)
   {
      if (dest==0x00) dest=0xFF; else dest--;

      if(dest==0x00) F|=Z_FLAG; else F&=(~Z_FLAG);
      F|=N_FLAG;
      if((dest&0xF)==0xF) F|=H_FLAG; else F&=(~H_FLAG);

      AF=combine(A,F);

      return dest;
   }

   /**
      The Decrease 16 bit op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy

      @param src the 16 bit source register
   **/

   private int DEC16(int dest)
   {
      if (dest==0x0000) dest=0xFFFF; else dest--;

      return dest;
   }

   /**
      The Increment op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy

      @param src the source register
   **/

   private char INC(char dest)
   {
      if (dest==0xFF) dest=0x00; else dest++;

      if(dest==0x00) F|=Z_FLAG; else F&=(~Z_FLAG);
      F&=(~N_FLAG);
      if((dest&0x0F)>0) F&=(~H_FLAG); else F|=H_FLAG;

      AF=combine(A,F);

      return dest;
   }

   /**
      The Increment 16 bit op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy

      @param src the source register
   **/

   private int INC16(int dest)
   {
      if (dest==0xFFFF) dest=0x0000; else dest++;

      return dest;
   }

   /**
      The jump op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy
   **/

   private void JP()
   {
      PC=getMemNN();
   }

   /**
      The jump relative op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy
   **/

   private void JR()
   {
      PC=getSignedMemN()+PC;
   }

   /**
      The or op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy

      @param src the source register
   **/

   private void OR(char src)
   {
      A|=src;
      A&=0xFF;

      F=(char)(A==0? Z_FLAG:0);

      AF=combine(A,F);
   }

   /**
      The POP op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy
   **/

   private int POP()
   {
      SP+=2;
      return (getMem(SP-1)<<8)|getMem(SP-2);
   }

   /**
      The PUSH op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy
   **/

   private void PUSH(int nn)
   {
      setMem(--SP, getUpper(nn));
      setMem(--SP, getLower(nn));
   }

   /**
      The res op code sets a specific bit in a register to 0. please refer to the opcode
      section of my report for a complete explanation of the op codes in a gameboy

      @param bit the bit to be set
      @param dest value to have the bit set
   **/

   private char RES(int bit, char dest)
   {
      return getLower(dest&(~(0x01<<bit)));
   }

   /**
      The Return op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy
   **/

   private void RET()
   {
      PC=POP();
   }

   /**
      The rotate left op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy

      @param dest the destination register
   **/

   private char RL(char dest)
   {
      tempOpCodeRegister=(dest<<1)|((F&C_FLAG)>0? 1:0);

      F=(char)(((tempOpCodeRegister&0xFF)==0? Z_FLAG:0) | (tempOpCodeRegister>0xFF? C_FLAG:0));

      AF=combine(A,F);
      return getLower(tempOpCodeRegister);
   }

   /**
      The rotate left acumulator op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy
   **/

   private void RLA()
   {
      tempOpCodeRegister=(A<<1)|((F&C_FLAG)>0? 1:0);

      F=(char)(tempOpCodeRegister>0xFF? C_FLAG:0);

      A=getLower(tempOpCodeRegister);
      AF=combine(A,F);
   }

   /**
      The rotate left with carry op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy

      @param dest the destination register
   **/

   private char RLC(char dest)
   {
      tempOpCodeRegister=(dest<<1)|BIT(dest,7);

      F=(char)(((tempOpCodeRegister&0xFF)==0? Z_FLAG:0) | (tempOpCodeRegister>0xFF? C_FLAG:0));

      AF=combine(A,F);
      return getLower(tempOpCodeRegister);
   }

   /**
      The rotate left with carry acumulator op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy
   **/

   private void RLCA()
   {
      tempOpCodeRegister=(A<<1)|BIT(A,7);

      F=(char)(tempOpCodeRegister>0xFF? C_FLAG:0);

      A=getLower(tempOpCodeRegister);
      AF=combine(A,F);
   }

   /**
      The rotate right op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy

      @param dest the destination register
   **/

   private char RR(char dest)
   {
      tempOpCodeRegister=(dest>>1)|(((F&C_FLAG)>0? 1:0)<<7);

      F=(char)((tempOpCodeRegister==0? Z_FLAG:0) | ((dest&0x01)>0? C_FLAG:0));

      AF=combine(A,F);
      return getLower(tempOpCodeRegister);
   }

   /**
      The rotate right acumulator op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy
   **/

   private void RRA()
   {
      tempOpCodeRegister=(A>>1)|(((F&C_FLAG)>0? 1:0)<<7);

      F=(char)((A&0x01)>0? C_FLAG:0);

      A=getLower(tempOpCodeRegister);
      AF=combine(A,F);
   }

   /**
      The rotate right with carry op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy

      @param dest the destination register
   **/

   private char RRC(char dest)
   {
      tempOpCodeRegister=(dest>>1)|(BIT(dest,0)<<7);

      F=(char)((tempOpCodeRegister==0? Z_FLAG:0) | ((dest&0x01)>0? C_FLAG:0));

      AF=combine(A,F);
      return getLower(tempOpCodeRegister);
   }

   /**
      The rotate right with carry acumulator op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy
   **/

   private void RRCA()
   {
      tempOpCodeRegister=(A>>1)|(BIT(A,0)<<7);

      F=(char)((A&0x01)>0? C_FLAG:0);

      A=getLower(tempOpCodeRegister);
      AF=combine(A,F);
   }

   /**
      The reset op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy

      @param b the position to jump to
   **/

   private void RST(int b)
   {
      PUSH(PC);
      PC=b;
   }

   /**
      The subtract with carry op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy

      @param src the source register
   **/

   private void SBC(char src)
   {
      tempOpCodeRegister=A-src-((F&C_FLAG)>0? 1:0);

      F=(char)(((tempOpCodeRegister&0xFF)==0? Z_FLAG:0) | N_FLAG | ((A&0x0F)<((src-((F&C_FLAG)>0? 1:0))&0x0F)? H_FLAG:0) | (tempOpCodeRegister<0? C_FLAG:0));

      A=getLower(tempOpCodeRegister);
      AF=combine(A,F);
   }

   /**
      The subtract op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy

      @param src the source register
   **/

   private void SUB(char src)
   {
      tempOpCodeRegister=A-src;

      F=(char)(((tempOpCodeRegister&0xFF)==0? Z_FLAG:0) | N_FLAG | ((A&0x0F)<(src&0x0F)? H_FLAG:0) | (tempOpCodeRegister<0? C_FLAG:0));

      A=getLower(tempOpCodeRegister);
      AF=combine(A,F);
   }

   /**
      The SCF op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy
   **/

   private void SCF()
   {
      F|=C_FLAG;
      F&=(~N_FLAG);
      F&=(~H_FLAG);

      AF=combine(A,F);
   }

   /**
      The set op code sets a specific bit in a register to 1. please refer to the opcode
      section of my report for a complete explanation of the op codes in a gameboy

      @param bit the bit to be set
      @param dest value to have the bit set
   **/

   private char SET(int bit, char dest)
   {
      return getLower(dest|(0x01<<bit));
   }

   /**
      The SLA op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy

      @param dest the destination register
   **/

   private char SLA(char dest)
   {
      tempOpCodeRegister=(dest<<1);

      F=(char)(((tempOpCodeRegister&0xFF)==0? Z_FLAG:0) | (tempOpCodeRegister>0xFF? C_FLAG:0));

      AF=combine(A,F);
      return getLower(tempOpCodeRegister);
   }

   /**
      The SRA op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy

      @param dest the destination register
   **/

   private char SRA(char dest)
   {
      tempOpCodeRegister=(dest>>1)|(dest&0x80);

      F=(char)((tempOpCodeRegister==0? Z_FLAG:0) | ((dest&0x01)>0? C_FLAG:0));

      AF=combine(A,F);
      return getLower(tempOpCodeRegister);
   }

   /**
      The SRL op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy

      @param dest the destination register
   **/

   private char SRL(char dest)
   {
      tempOpCodeRegister=(dest>>1);

      F=(char)((tempOpCodeRegister==0? Z_FLAG:0) | ((dest&0x01)>0? C_FLAG:0));

      AF=combine(A,F);
      return getLower(tempOpCodeRegister);
   }

   /**
      The SWAP op code swaps the upper and lower 4 bits around. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy

      @param dest the destination register
   **/

   private char SWAP(char dest)
   {
      F=(char)(dest==0? Z_FLAG:0);

      AF=combine(A,F);
      return SWAP_TABLE[dest];
   }

   /**
      The XOR op code. please refer to the opcode section of my report
      for a complete explanation of the op codes in a gameboy

      @param r the source register
   **/

   private void XOR(char r)
   {
      A^=r;
      A&=0xFF;

      F=(char)(A==0? Z_FLAG:0);

      AF=combine(A,F);
   }

   /**
      Increments important tally registers by a specified amount and performs some checks

      @param amount the value to increment by
   **/

   private void incAll(int amount)
   {
      CPU_CYCLES+=amount;
      DIV+=amount;
      if(DIV>(256<<CPU_SPEED))
      {
         DIV=0;
         GB_MEM[DIV_CNTR]++;

         if(GB_MEM[DIV_CNTR]==0xFF)
         {
            GB_MEM[DIV_CNTR]=0x00;
         }
      }
      if(TIMER_ON)
      {
         TIMER+=amount;
         if(TIMER>(TIMER_MAX<<CPU_SPEED))
         {
            TIMER=0;
            GB_MEM[TIMER_COUNT]++;

            if(GB_MEM[TIMER_COUNT]==0xFF)
            {
               GB_MEM[TIMER_COUNT]=0x00;//GB_MEM[TIMER_RELOAD];
               GB_MEM[INT_FLAG]|=0x04;
            }
         }
      }
   }

   /**
      The main method of the CPU. Calculates the mode logic, handles interrupts
      and works out which op code to run for one loop of the cpu
   **/

   public void opCodeList1()
   {
      //If the cpu has been stopped check to see if the cpu speed must be changed
      if (STOP) {if((GB_MEM[CPU_SPEED_REG]&0x01)>0) {GB_MEM[CPU_SPEED_REG]|=0x80; CPU_SPEED^=0x01; STOP=false;} return;}

      //Checks to see if its time to draw a line
      if(CPU_CYCLES>(456<<CPU_SPEED))
      {
         CPU_CYCLES=0x00;
         if(GB_MEM[LCD_Y_LOC]==0x99) GB_MEM[LCD_Y_LOC]=0x00; else GB_MEM[LCD_Y_LOC]++;

         //Checks the line compare to the current line
         if(GB_MEM[LCD_Y_LOC]==GB_MEM[LCD_Y_COMP])
         {
            GB_MEM[LCD_STAT]|=0x04;
            if((GB_MEM[LCD_STAT]&0x40)>0)
               GB_MEM[INT_FLAG]|=0x02;
         }
         else GB_MEM[LCD_STAT]&=0xFB;
      }

      //Mode selection code
      if(GB_MEM[LCD_Y_LOC]<0x90)
      {
         //Mode 10 logic
         if(CPU_CYCLES<(81<<CPU_SPEED) && (GB_MEM[LCD_STAT]&0x03)!=0x02)
         {
            GB_MEM[LCD_STAT]=(char)((GB_MEM[LCD_STAT]&0xFC)|0x02);
            if((GB_MEM[LCD_STAT]&0x20)>0)
               GB_MEM[INT_FLAG]|=0x02;
         }
         //Mode 11 logic
         else if(CPU_CYCLES>=(81<<CPU_SPEED) && CPU_CYCLES<((174+(12*SPR_PER_LINE))<<CPU_SPEED) && (GB_MEM[LCD_STAT]&0x03)!=0x03)
         {
            //Calculates the current line of graphics
            SPR_PER_LINE=VIDEO.drawLine();
            GB_MEM[LCD_STAT]=(char)((GB_MEM[LCD_STAT]&0xFC)|0x03);
         }
         //Mode 00 logic
         else if(CPU_CYCLES>=((174+(12*SPR_PER_LINE))<<CPU_SPEED) && (GB_MEM[LCD_STAT]&0x03)!=0)
         {
            GB_MEM[LCD_STAT]&=0xFC;
            if((GB_MEM[LCD_STAT]&0x08)>0)
               GB_MEM[INT_FLAG]|=0x02;
         }
      }
      //Mode 01 logic
      else if(GB_MEM[LCD_Y_LOC]>=0x90 && (GB_MEM[LCD_STAT]&0x03)!=0x01)
      {
         //Draws a full screen of graphics to the screen
         VIDEO.refresh();
         GB_MEM[LCD_STAT]=(char)((GB_MEM[LCD_STAT]&0xFC)|0x01);
         if((GB_MEM[LCD_STAT]&0x10)>0)
            GB_MEM[INT_FLAG]|=0x02;
         GB_MEM[INT_FLAG]|=0x01;
      }

      //Interrupt code
      if ((IME)&&(GB_MEM[INT_FLAG]>0))
      {
         if      (((GB_MEM[INT_FLAG]&0x01)>0)&&((GB_MEM[INT_ENABLE]&0x01)>0)) {HALT=false; PUSH(PC); PC=0x40; IME=false; GB_MEM[INT_FLAG]&=(~0x01);} //V-Blank
         else if (((GB_MEM[INT_FLAG]&0x02)>0)&&((GB_MEM[INT_ENABLE]&0x02)>0)) {HALT=false; PUSH(PC); PC=0x48; IME=false; GB_MEM[INT_FLAG]&=(~0x02);} //LCDC
         else if (((GB_MEM[INT_FLAG]&0x04)>0)&&((GB_MEM[INT_ENABLE]&0x04)>0)) {HALT=false; PUSH(PC); PC=0x50; IME=false; GB_MEM[INT_FLAG]&=(~0x04);} //Timer overflow
         else if (((GB_MEM[INT_FLAG]&0x08)>0)&&((GB_MEM[INT_ENABLE]&0x08)>0)) {HALT=false; PUSH(PC); PC=0x58; IME=false; GB_MEM[INT_FLAG]&=(~0x08);} //Serial
         else if (((GB_MEM[INT_FLAG]&0x10)>0)&&((GB_MEM[INT_ENABLE]&0x10)>0)) {HALT=false; PUSH(PC); PC=0x60; IME=false; GB_MEM[INT_FLAG]&=(~0x10);} //Hi->lo
      }

      //Error checking to see that the PC doesnt go out of memory
      if (PC>0xFFFF) {GB_MEM[CPU_SPEED_REG]=0x00; STOP=true;  System.out.print("   Program Counter out of Memory!!!\n"); return;}

      if (HALT) {incAll(4); return;}

      //gets the next op code
      opCode=getMem(PC++);
      incAll(incAllValue[opCode]);

      //Works out which op code to run a runs it
      switch (opCode)
      {
         case 0x00 : {                                                          break;} //NOP
         case 0x01 : {BC=getMemNN(); B=getUpper(BC); C=getLower(BC);          break;} //LD BC,$aabb
         case 0x02 : {setMem(BC, A);                                           break;} //LD (BC),A
         case 0x03 : {BC=INC16(BC); B=getUpper(BC); C=getLower(BC);                       break;} //INC BC
         case 0x04 : {B=INC(B); BC=combine(B,C);                                          break;} //INC B
         case 0x05 : {B=DEC(B); BC=combine(B,C);                                          break;} //DEC B
         case 0x06 : {B=getMem(PC++); BC=combine(B,C);                            break;} //LD B,$xx
         case 0x07 : {RLCA();                                                             break;} //RLCA
         case 0x08 : {int tempLoc=getMemNN(); setMem(tempLoc, getLower(SP));  setMem(tempLoc+1, getUpper(SP)); break;} //LD ($aabb),SP
         case 0x09 : {HL=ADD16(HL, BC); H=getUpper(HL); L=getLower(HL);                   break;} //ADD HL,BC
         case 0x0A : {A=getMem(BC); AF=combine(A,F);                           break;} //LD A,(BC)
         case 0x0B : {BC=DEC16(BC); B=getUpper(BC); C=getLower(BC);                       break;} //DEC BC
         case 0x0C : {C=INC(C); BC=combine(B,C);                                          break;} //INC C
         case 0x0D : {C=DEC(C); BC=combine(B,C);                                          break;} //DEC C
         case 0x0E : {C=getMem(PC++); BC=combine(B,C);                            break;} //LD C,$xx
         case 0x0F : {RRCA();                                                             break;} //RRCA

         case 0x10 : {STOP=true;                                               break;} //STOP
         case 0x11 : {DE=getMemNN(); D=getUpper(DE); E=getLower(DE);          break;} //LD DE,$aabb
         case 0x12 : {setMem(DE, A);                                           break;} //LD (DE),A
         case 0x13 : {DE=INC16(DE); D=getUpper(DE); E=getLower(DE);                       break;} //INC DE
         case 0x14 : {D=INC(D); DE=combine(D,E);                                          break;} //INC D
         case 0x15 : {D=DEC(D); DE=combine(D,E);                                          break;} //DEC D
         case 0x16 : {D=getMem(PC++); DE=combine(D,E);                            break;} //LD D,$xx
         case 0x17 : {RLA();                                                              break;} //RLA
         case 0x18 : {JR();                                                               break;} //JR $xx
         case 0x19 : {HL=ADD16(HL, DE); H=getUpper(HL); L=getLower(HL);                   break;} //ADD HL,DE
         case 0x1A : {A=getMem(DE); AF=combine(A,F);                           break;} //LD A,(DE)
         case 0x1B : {DE=DEC16(DE); D=getUpper(DE); E=getLower(DE);                       break;} //DEC DE
         case 0x1C : {E=INC(E); DE=combine(D,E);                                          break;} //INC E
         case 0x1D : {E=DEC(E); DE=combine(D,E);                                          break;} //DEC E
         case 0x1E : {E=getMem(PC++); DE=combine(D,E);                            break;} //LD E,$xx
         case 0x1F : {RRA();                                                              break;} //RRA

         case 0x20 : {if((F&Z_FLAG)==0x00) {incAll(4); JR();} else {PC++;}                break;} //JR NZ,$xx
         case 0x21 : {HL=getMemNN(); H=getUpper(HL); L=getLower(HL);          break;} //LD HL,$aabb
         case 0x22 : {setMem(HL, A); HL=INC16(HL); H=getUpper(HL); L=getLower(HL);        break;} //LD (HLI),A
         case 0x23 : {HL=INC16(HL); H=getUpper(HL); L=getLower(HL);                       break;} //INC HL
         case 0x24 : {H=INC(H); HL=combine(H,L);                                          break;} //INC H
         case 0x25 : {H=DEC(H); HL=combine(H,L);                                          break;} //DEC H
         case 0x26 : {H=getMem(PC++); HL=combine(H,L);                            break;} //LD H,$xx
         case 0x27 : {DAA();                                                              break;} //DAA
         case 0x28 : {if((F&Z_FLAG)>0) {incAll(4); JR();} else {PC++;}                    break;} //JR Z,$xx
         case 0x29 : {HL=ADD16(HL,HL); H=getUpper(HL); L=getLower(HL);                    break;} //ADD HL,HL
         case 0x2A : {A=getMem(HL); AF=combine(A,F); HL=INC16(HL); H=getUpper(HL); L=getLower(HL); break;} //LD A,(HLI)
         case 0x2B : {HL=DEC16(HL); H=getUpper(HL); L=getLower(HL);                       break;} //DEC HL
         case 0x2C : {L=INC(L); HL=combine(H,L);                                          break;} //INC L
         case 0x2D : {L=DEC(L); HL=combine(H,L);                                          break;} //DEC L
         case 0x2E : {L=getMem(PC++); HL=combine(H,L);                            break;} //LD L,$xx
         case 0x2F : {CPL();                                                              break;} //CPL

         case 0x30 : {if((F&C_FLAG)==0x00) {incAll(4); JR();} else {PC++;}                break;} //JR NC,$xx
         case 0x31 : {SP=getMemNN();                                          break;} //SP,$aabb
         case 0x32 : {setMem(HL, A); HL=DEC16(HL); H=getUpper(HL); L=getLower(HL);        break;} //LD (HLD),A
         case 0x33 : {SP=INC16(SP);                                                       break;} //INC SP
         case 0x34 : {setMem(HL, INC(getMem(HL)));                             break;} //INC (HL)
         case 0x35 : {setMem(HL, DEC(getMem(HL)));                             break;} //DEC (HL)
         case 0x36 : {setMem(HL, getMem(PC++));                                  break;} //LD (HL),$xx
         case 0x37 : {SCF();                                                              break;} //SCF
         case 0x38 : {if((F&C_FLAG)>0) {incAll(4); JR();} else {PC++;}                    break;} //JR C,$xx
         case 0x39 : {HL=ADD16(HL,SP); H=getUpper(HL); L=getLower(HL);                    break;} //ADD HL,SP
         case 0x3A : {A=getMem(HL); AF=combine(A,F); HL=DEC16(HL); H=getUpper(HL); L=getLower(HL); break;} //LD A,(HLD),
         case 0x3B : {SP=DEC16(SP);                                                       break;} //DEC SP
         case 0x3C : {A=INC(A); AF=combine(A,F);                                          break;} //INC A
         case 0x3D : {A=DEC(A); AF=combine(A,F);                                          break;} //DEC A
         case 0x3E : {A=getMem(PC++); AF=combine(A,F);                            break;} //LD A,$xx
         case 0x3F : {CCF();                                                              break;} //CCF

         case 0x40 : {B=B; BC=combine(B,C);                    break;} //LD B,B
         case 0x41 : {B=C; BC=combine(B,C);                    break;} //LD B,C
         case 0x42 : {B=D; BC=combine(B,C);                    break;} //LD B,D
         case 0x43 : {B=E; BC=combine(B,C);                    break;} //LD B,E
         case 0x44 : {B=H; BC=combine(B,C);                    break;} //LD B,H
         case 0x45 : {B=L; BC=combine(B,C);                    break;} //LD B,L
         case 0x46 : {B=getMem(HL); BC=combine(B,C);           break;} //LD B,(HL)
         case 0x47 : {B=A; BC=combine(B,C);                    break;} //LD B,A
         case 0x48 : {C=B; BC=combine(B,C);                    break;} //LD C,B
         case 0x49 : {C=C; BC=combine(B,C);                    break;} //LD C,C
         case 0x4A : {C=D; BC=combine(B,C);                    break;} //LD C,D
         case 0x4B : {C=E; BC=combine(B,C);                    break;} //LD C,E
         case 0x4C : {C=H; BC=combine(B,C);                    break;} //LD C,H
         case 0x4D : {C=L; BC=combine(B,C);                    break;} //LD C,L
         case 0x4E : {C=getMem(HL); BC=combine(B,C);           break;} //LD C,(HL)
         case 0x4F : {C=A; BC=combine(B,C);                    break;} //LD C,A

         case 0x50 : {D=B; DE=combine(D,E);                    break;} //LD D,B
         case 0x51 : {D=C; DE=combine(D,E);                    break;} //LD D,C
         case 0x52 : {D=D; DE=combine(D,E);                    break;} //LD D,D
         case 0x53 : {D=E; DE=combine(D,E);                    break;} //LD D,E
         case 0x54 : {D=H; DE=combine(D,E);                    break;} //LD D,H
         case 0x55 : {D=L; DE=combine(D,E);                    break;} //LD D,L
         case 0x56 : {D=getMem(HL); DE=combine(D,E);           break;} //LD D,(HL)
         case 0x57 : {D=A; DE=combine(D,E);                    break;} //LD D,A
         case 0x58 : {E=B; DE=combine(D,E);                    break;} //LD E,B
         case 0x59 : {E=C; DE=combine(D,E);                    break;} //LD E,C
         case 0x5A : {E=D; DE=combine(D,E);                    break;} //LD E,D
         case 0x5B : {E=E; DE=combine(D,E);                    break;} //LD E,E
         case 0x5C : {E=H; DE=combine(D,E);                    break;} //LD E,H
         case 0x5D : {E=L; DE=combine(D,E);                    break;} //LD E,L
         case 0x5E : {E=getMem(HL); DE=combine(D,E);           break;} //LD E,(HL)
         case 0x5F : {E=A; DE=combine(D,E);                    break;} //LD E,A

         case 0x60 : {H=B; HL=combine(H,L);                    break;} //LD H,B
         case 0x61 : {H=C; HL=combine(H,L);                    break;} //LD H,C
         case 0x62 : {H=D; HL=combine(H,L);                    break;} //LD H,D
         case 0x63 : {H=E; HL=combine(H,L);                    break;} //LD H,E
         case 0x64 : {H=H; HL=combine(H,L);                    break;} //LD H,H
         case 0x65 : {H=L; HL=combine(H,L);                    break;} //LD H,L
         case 0x66 : {H=getMem(HL); HL=combine(H,L);           break;} //LD H,(HL)
         case 0x67 : {H=A; HL=combine(H,L);                    break;} //LD H,A
         case 0x68 : {L=B; HL=combine(H,L);                    break;} //LD L,B
         case 0x69 : {L=C; HL=combine(H,L);                    break;} //LD L,C
         case 0x6A : {L=D; HL=combine(H,L);                    break;} //LD L,D
         case 0x6B : {L=E; HL=combine(H,L);                    break;} //LD L,E
         case 0x6C : {L=H; HL=combine(H,L);                    break;} //LD L,H
         case 0x6D : {L=L; HL=combine(H,L);                    break;} //LD L,L
         case 0x6E : {L=getMem(HL); HL=combine(H,L);           break;} //LD L,(HL)
         case 0x6F : {L=A; HL=combine(H,L);                    break;} //LD L,A

         case 0x70 : {setMem(HL, B);                           break;} //LD (HL),B
         case 0x71 : {setMem(HL, C);                           break;} //LD (HL),C
         case 0x72 : {setMem(HL, D);                           break;} //LD (HL),D
         case 0x73 : {setMem(HL, E);                           break;} //LD (HL),E
         case 0x74 : {setMem(HL, H);                           break;} //LD (HL),H
         case 0x75 : {setMem(HL, L);                           break;} //LD (HL),L
         case 0x76 : {if(IME) HALT=true; else HALT_NEXT=true;  break;} //HALT
         case 0x77 : {setMem(HL, A);                           break;} //LD (HL),A
         case 0x78 : {A=B; AF=combine(A,F);                    break;} //LD A,B
         case 0x79 : {A=C; AF=combine(A,F);                    break;} //LD A,C
         case 0x7A : {A=D; AF=combine(A,F);                    break;} //LD A,D
         case 0x7B : {A=E; AF=combine(A,F);                    break;} //LD A,E
         case 0x7C : {A=H; AF=combine(A,F);                    break;} //LD A,H
         case 0x7D : {A=L; AF=combine(A,F);                    break;} //LD A,L
         case 0x7E : {A=getMem(HL); AF=combine(A,F);           break;} //LD A,(HL)
         case 0x7F : {A=A; AF=combine(A,F);                    break;} //LD A,A

         case 0x80 : {ADD(B);                                   break;} //ADD A,B
         case 0x81 : {ADD(C);                                   break;} //ADD A,C
         case 0x82 : {ADD(D);                                   break;} //ADD A,D
         case 0x83 : {ADD(E);                                   break;} //ADD A,E
         case 0x84 : {ADD(H);                                   break;} //ADD A,H
         case 0x85 : {ADD(L);                                   break;} //ADD A,L
         case 0x86 : {ADD(getMem(HL));                          break;} //ADD A,(HL)
         case 0x87 : {ADD(A);                                   break;} //ADD A,A
         case 0x88 : {ADC(B);                                  break;} //ADDC A,B
         case 0x89 : {ADC(C);                                  break;} //ADDC A,C
         case 0x8A : {ADC(D);                                  break;} //ADDC A,D
         case 0x8B : {ADC(E);                                  break;} //ADDC A,E
         case 0x8C : {ADC(H);                                  break;} //ADDC A,H
         case 0x8D : {ADC(L);                                  break;} //ADDC A,L
         case 0x8E : {ADC(getMem(HL));                         break;} //ADDC A,(HL)
         case 0x8F : {ADC(A);                                  break;} //ADDC A,A

         case 0x90 : {SUB(B);                                   break;} //SUB B
         case 0x91 : {SUB(C);                                   break;} //SUB C
         case 0x92 : {SUB(D);                                   break;} //SUB D
         case 0x93 : {SUB(E);                                   break;} //SUB E
         case 0x94 : {SUB(H);                                   break;} //SUB H
         case 0x95 : {SUB(L);                                   break;} //SUB L
         case 0x96 : {SUB(getMem(HL));                          break;} //SUB (HL)
         case 0x97 : {SUB(A);                                   break;} //SUB A
         case 0x98 : {SBC(B);                                  break;} //SUBC B
         case 0x99 : {SBC(C);                                  break;} //SUBC C
         case 0x9A : {SBC(D);                                  break;} //SUBC D
         case 0x9B : {SBC(E);                                  break;} //SUBC E
         case 0x9C : {SBC(H);                                  break;} //SUBC H
         case 0x9D : {SBC(L);                                  break;} //SUBC L
         case 0x9E : {SBC(getMem(HL));                         break;} //SUBC (HL)
         case 0x9F : {SBC(A);                                  break;} //SUBC A

         case 0xA0 : {AND(B);                                   break;} //AND B
         case 0xA1 : {AND(C);                                   break;} //AND C
         case 0xA2 : {AND(D);                                   break;} //AND D
         case 0xA3 : {AND(E);                                   break;} //AND E
         case 0xA4 : {AND(H);                                   break;} //AND H
         case 0xA5 : {AND(L);                                   break;} //AND L
         case 0xA6 : {AND(getMem(HL));                          break;} //AND (HL)
         case 0xA7 : {AND(A);                                   break;} //AND A
         case 0xA8 : {XOR(B);                                   break;} //XOR B
         case 0xA9 : {XOR(C);                                   break;} //XOR C
         case 0xAA : {XOR(D);                                   break;} //XOR D
         case 0xAB : {XOR(E);                                   break;} //XOR E
         case 0xAC : {XOR(H);                                   break;} //XOR H
         case 0xAD : {XOR(L);                                   break;} //XOR L
         case 0xAE : {XOR(getMem(HL));                          break;} //XOR (HL)
         case 0xAF : {XOR(A);                                   break;} //XOR A

         case 0xB0 : {OR(B);                                    break;} //OR B
         case 0xB1 : {OR(C);                                    break;} //OR C
         case 0xB2 : {OR(D);                                    break;} //OR D
         case 0xB3 : {OR(E);                                    break;} //OR E
         case 0xB4 : {OR(H);                                    break;} //OR H
         case 0xB5 : {OR(L);                                    break;} //OR L
         case 0xB6 : {OR(getMem(HL));                           break;} //OR (HL)
         case 0xB7 : {OR(A);                                    break;} //OR A
         case 0xB8 : {CP(B);                                    break;} //CP B
         case 0xB9 : {CP(C);                                    break;} //CP C
         case 0xBA : {CP(D);                                    break;} //CP D
         case 0xBB : {CP(E);                                    break;} //CP E
         case 0xBC : {CP(H);                                    break;} //CP H
         case 0xBD : {CP(L);                                    break;} //CP L
         case 0xBE : {CP(getMem(HL));                           break;} //CP (HL)
         case 0xBF : {CP(A);                                    break;} //CP A

         case 0xC0 : {if((F&Z_FLAG)==0){incAll(12); RET();}                            break;} //RET NZ
         case 0xC1 : {BC=POP(); B=getUpper(BC); C=getLower(BC);                        break;} //POP BC
         case 0xC2 : {if((F&Z_FLAG)==0){incAll(4); JP();} else PC+=2;                           break;} //JP NZ,$aabb
         case 0xC3 : {JP();                                                            break;} //JP $aabb
         case 0xC4 : {if((F&Z_FLAG)==0){incAll(12); CALL();} else PC+=2;             break;} //CALL NZ,$aabb
         case 0xC5 : {PUSH(BC);                                                        break;} //PUSH BC
         case 0xC6 : {ADD(getMem(PC++));                                       break;} //ADD A,$xx
         case 0xC7 : {RST(0x00);                                                       break;} //RST $00
         case 0xC8 : {if((F&Z_FLAG)==Z_FLAG){incAll(12); RET();}                       break;} //RET Z
         case 0xC9 : {RET();                                                           break;} //RET
         case 0xCA : {if((F&Z_FLAG)==Z_FLAG){incAll(4); JP();} else PC+=2;                      break;} //JP Z,$aabb
         case 0xCB : {singleCPUStepCB();                                               break;} //CB Oppcodes
         case 0xCC : {if((F&Z_FLAG)==Z_FLAG){incAll(12); CALL();} else PC+=2;        break;} // bb aa CALL Z,$aabb",
         case 0xCD : {CALL();                                                          break;} //CALL nn
         case 0xCE : {ADC(getMem(PC++));                                       break;} //ADC A,$xx
         case 0xCF : {RST(0x08);                                                       break;} //RST $08
   
         case 0xD0 : {if((F&C_FLAG)==0){incAll(12); RET();}                            break;} //RET NC
         case 0xD1 : {DE=POP(); D=getUpper(DE); E=getLower(DE);                        break;} //POP DE
         case 0xD2 : {if((F&C_FLAG)==0){incAll(4); JP();} else PC+=2;                           break;} //JP NC,$aabb
         case 0xD3 : {GB_MEM[CPU_SPEED_REG]=0x00; STOP=true; System.out.println("Un needed opCode 0xD3 "+Integer.toHexString(PC-1)); break;} //NOP
         case 0xD4 : {if((F&C_FLAG)==0){incAll(12); CALL();} else PC+=2;             break;} //CALL NC,$aabb
         case 0xD5 : {PUSH(DE);                                                        break;} //PUSH DE
         case 0xD6 : {SUB(getMem(PC++));                                       break;} //SUB $xx
         case 0xD7 : {RST(0x10);                                                       break;} //RST $10
         case 0xD8 : {if((F&C_FLAG)==C_FLAG){incAll(12); RET();}                       break;} //RET C
         case 0xD9 : {RET(); opCodeList1(); IME=true; if(HALT_NEXT) {HALT=true; HALT_NEXT=false;} break;} //RETI
         case 0xDA : {if((F&C_FLAG)==C_FLAG){incAll(4); JP();} else PC+=2;                      break;} //JP C,$aabb
         case 0xDB : {GB_MEM[CPU_SPEED_REG]=0x00; STOP=true; System.out.println("Un needed opCode 0xDB "+Integer.toHexString(PC-1)); break;} //NOP
         case 0xDC : {if((F&C_FLAG)==C_FLAG){incAll(12); CALL();} else PC+=2;        break;} // bb aa CALL C,$aabb",
         case 0xDD : {GB_MEM[CPU_SPEED_REG]=0x00; STOP=true; System.out.println("Un needed opCode 0xDD "+Integer.toHexString(PC-1)); break;} //NOP
         case 0xDE : {SBC(getMem(PC++));                                       break;} //SBC A,$xx
         case 0xDF : {RST(0x18);                                                       break;} //RST $18

         case 0xE0 : {setMem(0xFF00+getMem(PC++), A);                         break;} //LD ($FFxx),A
         case 0xE1 : {HL=POP(); H=getUpper(HL); L=getLower(HL);                        break;} //POP HL
         case 0xE2 : {setMem(0xFF00+C, A);                                 break;} //LD ($FF00+C),A
         case 0xE3 : {GB_MEM[CPU_SPEED_REG]=0x00; STOP=true; System.out.println("Un needed opCode 0xE3 "+Integer.toHexString(PC-1)); break;} //NOP
         case 0xE4 : {GB_MEM[CPU_SPEED_REG]=0x00; STOP=true; System.out.println("Un needed opCode 0xE4 "+Integer.toHexString(PC-1)); break;} //NOP
         case 0xE5 : {PUSH(HL);                                                        break;} //PUSH HL
         case 0xE6 : {AND(getMem(PC++));                                       break;} //AND $xx
         case 0xE7 : {RST(0x20);                                                       break;} //RST 0x20
         case 0xE8 : {SP=ADDSP(getSignedMemN()); break;} //ADD SP,xx
         case 0xE9 : {PC=HL;                                                break;} //JP (HL)
         case 0xEA : {setMem(getMemNN(), A);                               break;} //LD ($aabb),A
         case 0xEB : {GB_MEM[CPU_SPEED_REG]=0x00; STOP=true; System.out.println("Un needed opCode 0xEB "+Integer.toHexString(PC-1)); break;} //NOP
         case 0xEC : {GB_MEM[CPU_SPEED_REG]=0x00; STOP=true; System.out.println("Un needed opCode 0xEC "+Integer.toHexString(PC-1)); break;} //NOP
         case 0xED : {GB_MEM[CPU_SPEED_REG]=0x00; STOP=true; System.out.println("Un needed opCode 0xED "+Integer.toHexString(PC-1)); break;} //ED Oppcodes or NOP
         case 0xEE : {XOR(getMem(PC++));                                       break;} //XOR $xx
         case 0xEF : {RST(0x28);                                                       break;} //RST 0x28

         case 0xF0 : {A=getMem(0xFF00+getMem(PC++)); AF=combine(A,F);         break;} //LD A,($FFxx)
         case 0xF1 : {AF=POP(); A=getUpper(AF); F=getLower(AF);                        break;} //POP AF
         case 0xF2 : {A=getMem(0xFF00+C); AF=combine(A,F);                 break;} //LD A,($FF00+C)
         case 0xF3 : {IME=false;                                            break;} //DI
         case 0xF4 : {GB_MEM[CPU_SPEED_REG]=0x00; STOP=true; System.out.println("Un needed opCode 0xF4 "+Integer.toHexString(PC-1)); break;} //NOP
         case 0xF5 : {PUSH(AF);                                                        break;} //PUSH AF
         case 0xF6 : {OR(getMem(PC++));                                        break;} //OR $xx
         case 0xF7 : {RST(0x30);                                                       break;} //RST 0x30
         case 0xF8 : {HL=ADDSP(getSignedMemN()); H=getUpper(HL); L=getLower(HL); break;} //LD HL,SP+e
         case 0xF9 : {SP=HL;                                                break;} //LD SP,HL
         case 0xFA : {A=getMem(getMemNN()); AF=combine(A,F);                break;} //LD A (nn)
         case 0xFB : {opCodeList1(); IME=true; if(HALT_NEXT) {HALT=true; HALT_NEXT=false;} break;} //EI
         case 0xFC : {GB_MEM[CPU_SPEED_REG]=0x00; STOP=true; System.out.println("Un needed opCode 0xFC "+Integer.toHexString(PC-1)); break;} //NOP
         case 0xFD : {GB_MEM[CPU_SPEED_REG]=0x00; STOP=true; System.out.println("Un needed opCode 0xFD "+Integer.toHexString(PC-1)); break;} //NOP
         case 0xFE : {CP(getMem(PC++));                                        break;} //CP N
         case 0xFF : {RST(0x38);                                                       break;} //RST 0x38

         default : {System.out.println("Unknown op code " + Integer.toHexString(opCode)); break;}
      }
   }

   /**
      The CB or secondary opcodes
   **/

   private void singleCPUStepCB()
   {
      opCode2=getMem(PC++);
      incAll(incAllValue[opCode2]);

      //Works out which op code to run a runs it
      switch (opCode2)
      {
         case 0x00 : {B=RLC(B); BC=combine(B,C);                break;} //RLC B
         case 0x01 : {C=RLC(C); BC=combine(B,C);                break;} //RLC C
         case 0x02 : {D=RLC(D); DE=combine(D,E);                break;} //RLC D
         case 0x03 : {E=RLC(E); DE=combine(D,E);                break;} //RLC E
         case 0x04 : {H=RLC(H); HL=combine(H,L);                break;} //RLC H
         case 0x05 : {L=RLC(L); HL=combine(H,L);                break;} //RLC L
         case 0x06 : {setMem(HL, RLC(getMem(HL)));             break;} //RLC (HL)
         case 0x07 : {A=RLC(A); AF=combine(A,F);                break;} //RLC A
         case 0x08 : {B=RRC(B); BC=combine(B,C);                break;} //RRC B
         case 0x09 : {C=RRC(C); BC=combine(B,C);                break;} //RRC C
         case 0x0A : {D=RRC(D); DE=combine(D,E);                break;} //RRC D
         case 0x0B : {E=RRC(E); DE=combine(D,E);                break;} //RRC E
         case 0x0C : {H=RRC(H); HL=combine(H,L);                break;} //RRC H
         case 0x0D : {L=RRC(L); HL=combine(H,L);                break;} //RRC L
         case 0x0E : {setMem(HL, RRC(getMem(HL)));             break;} //RRC (HL)
         case 0x0F : {A=RRC(A); AF=combine(A,F);                break;} //RRC A

         case 0x10 : {B=RL(B); BC=combine(B,C);                 break;} //RL B
         case 0x11 : {C=RL(C); BC=combine(B,C);                 break;} //RL C
         case 0x12 : {D=RL(D); DE=combine(D,E);                 break;} //RL D
         case 0x13 : {E=RL(E); DE=combine(D,E);                 break;} //RL E
         case 0x14 : {H=RL(H); HL=combine(H,L);                 break;} //RL H
         case 0x15 : {L=RL(L); HL=combine(H,L);                 break;} //RL L
         case 0x16 : {setMem(HL, RL(getMem(HL)));              break;} //RL (HL)
         case 0x17 : {A=RL(A); AF=combine(A,F);                 break;} //RL A
         case 0x18 : {B=RR(B); BC=combine(B,C);                 break;} //RR B
         case 0x19 : {C=RR(C); BC=combine(B,C);                 break;} //RR C
         case 0x1A : {D=RR(D); DE=combine(D,E);                 break;} //RR D
         case 0x1B : {E=RR(E); DE=combine(D,E);                 break;} //RR E
         case 0x1C : {H=RR(H); HL=combine(H,L);                 break;} //RR H
         case 0x1D : {L=RR(L); HL=combine(H,L);                 break;} //RR L
         case 0x1E : {setMem(HL, RR(getMem(HL)));              break;} //RR (HL)
         case 0x1F : {A=RR(A); AF=combine(A,F);                 break;} //RR A

         case 0x20 : {B=SLA(B); BC=combine(B,C);                break;} //SLA B
         case 0x21 : {C=SLA(C); BC=combine(B,C);                break;} //SLA C
         case 0x22 : {D=SLA(D); DE=combine(D,E);                break;} //SLA D
         case 0x23 : {E=SLA(E); DE=combine(D,E);                break;} //SLA E
         case 0x24 : {H=SLA(H); HL=combine(H,L);                break;} //SLA H
         case 0x25 : {L=SLA(L); HL=combine(H,L);                break;} //SLA L
         case 0x26 : {setMem(HL, SLA(getMem(HL)));             break;} //SLA (HL)
         case 0x27 : {A=SLA(A); AF=combine(A,F);                break;} //SLA A
         case 0x28 : {B=SRA(B); BC=combine(B,C);                break;} //SRA B
         case 0x29 : {C=SRA(C); BC=combine(B,C);                break;} //SRA C
         case 0x2A : {D=SRA(D); DE=combine(D,E);                break;} //SRA D
         case 0x2B : {E=SRA(E); DE=combine(D,E);                break;} //SRA E
         case 0x2C : {H=SRA(H); HL=combine(H,L);                break;} //SRA H
         case 0x2D : {L=SRA(L); HL=combine(H,L);                break;} //SRA L
         case 0x2E : {setMem(HL, SRA(getMem(HL)));             break;} //SRA (HL)
         case 0x2F : {A=SRA(A); AF=combine(A,F);                break;} //SRA A

         case 0x30 : {B=SWAP(B); BC=combine(B,C);               break;} //
         case 0x31 : {C=SWAP(C); BC=combine(B,C);               break;} //
         case 0x32 : {D=SWAP(D); DE=combine(D,E);               break;} //
         case 0x33 : {E=SWAP(E); DE=combine(D,E);               break;} //
         case 0x34 : {H=SWAP(H); HL=combine(H,L);               break;} //
         case 0x35 : {L=SWAP(L); HL=combine(H,L);               break;} //
         case 0x36 : {setMem(HL,SWAP(getMem(HL)));              break;} //
         case 0x37 : {A=SWAP(A); AF=combine(A,F);               break;} //
         case 0x38 : {B=SRL(B); BC=combine(B,C);                break;} //SLA B
         case 0x39 : {C=SRL(C); BC=combine(B,C);                break;} //SLA C
         case 0x3A : {D=SRL(D); DE=combine(D,E);                break;} //SLA D
         case 0x3B : {E=SRL(E); DE=combine(D,E);                break;} //SLA E
         case 0x3C : {H=SRL(H); HL=combine(H,L);                break;} //SLA H
         case 0x3D : {L=SRL(L); HL=combine(H,L);                break;} //SLA L
         case 0x3E : {setMem(HL, SRL(getMem(HL)));             break;} //SLA (HL)
         case 0x3F : {A=SRL(A); AF=combine(A,F);                break;} //SLA A

         case 0x40 : {BIT(0,B);                                 break;} //BIT 0,B
         case 0x41 : {BIT(0,C);                                 break;} //BIT 0,C
         case 0x42 : {BIT(0,D);                                 break;} //BIT 0,D
         case 0x43 : {BIT(0,E);                                 break;} //BIT 0,E
         case 0x44 : {BIT(0,H);                                 break;} //BIT 0,H
         case 0x45 : {BIT(0,L);                                 break;} //BIT 0,L
         case 0x46 : {BIT(0, getMem(HL));                      break;} //BIT 0,(HL)
         case 0x47 : {BIT(0,A);                                 break;} //BIT 0,A
         case 0x48 : {BIT(1,B);                                 break;} //BIT 1,B
         case 0x49 : {BIT(1,C);                                 break;} //BIT 1,C
         case 0x4A : {BIT(1,D);                                 break;} //BIT 1,D
         case 0x4B : {BIT(1,E);                                 break;} //BIT 1,E
         case 0x4C : {BIT(1,H);                                 break;} //BIT 1,H
         case 0x4D : {BIT(1,L);                                 break;} //BIT 1,L
         case 0x4E : {BIT(1, getMem(HL));                      break;} //BIT 1,(HL)
         case 0x4F : {BIT(1,A);                                 break;} //BIT 1,A

         case 0x50 : {BIT(2,B);                                 break;} //BIT 2,B
         case 0x51 : {BIT(2,C);                                 break;} //BIT 2,C
         case 0x52 : {BIT(2,D);                                 break;} //BIT 2,D
         case 0x53 : {BIT(2,E);                                 break;} //BIT 2,E
         case 0x54 : {BIT(2,H);                                 break;} //BIT 2,H
         case 0x55 : {BIT(2,L);                                 break;} //BIT 2,L
         case 0x56 : {BIT(2, getMem(HL));                      break;} //BIT 2,(HL)
         case 0x57 : {BIT(2,A);                                 break;} //BIT 2,A
         case 0x58 : {BIT(3,B);                                 break;} //BIT 3,B
         case 0x59 : {BIT(3,C);                                 break;} //BIT 3,C
         case 0x5A : {BIT(3,D);                                 break;} //BIT 3,D
         case 0x5B : {BIT(3,E);                                 break;} //BIT 3,E
         case 0x5C : {BIT(3,H);                                 break;} //BIT 3,H
         case 0x5D : {BIT(3,L);                                 break;} //BIT 3,L
         case 0x5E : {BIT(3, getMem(HL));                      break;} //BIT 3,(HL)
         case 0x5F : {BIT(3,A);                                 break;} //BIT 3,A

         case 0x60 : {BIT(4,B);                                 break;} //BIT 4,B
         case 0x61 : {BIT(4,C);                                 break;} //BIT 4,C
         case 0x62 : {BIT(4,D);                                 break;} //BIT 4,D
         case 0x63 : {BIT(4,E);                                 break;} //BIT 4,E
         case 0x64 : {BIT(4,H);                                 break;} //BIT 4,H
         case 0x65 : {BIT(4,L);                                 break;} //BIT 4,L
         case 0x66 : {BIT(4, getMem(HL));                      break;} //BIT 4,(HL)
         case 0x67 : {BIT(4,A);                                 break;} //BIT 4,A
         case 0x68 : {BIT(5,B);                                 break;} //BIT 5,B
         case 0x69 : {BIT(5,C);                                 break;} //BIT 5,C
         case 0x6A : {BIT(5,D);                                 break;} //BIT 5,D
         case 0x6B : {BIT(5,E);                                 break;} //BIT 5,E
         case 0x6C : {BIT(5,H);                                 break;} //BIT 5,H
         case 0x6D : {BIT(5,L);                                 break;} //BIT 5,L
         case 0x6E : {BIT(5, getMem(HL));                      break;} //BIT 5,(HL)
         case 0x6F : {BIT(5,A);                                 break;} //BIT 5,A

         case 0x70 : {BIT(6,B);                                 break;} //BIT 6,B
         case 0x71 : {BIT(6,C);                                 break;} //BIT 6,C
         case 0x72 : {BIT(6,D);                                 break;} //BIT 6,D
         case 0x73 : {BIT(6,E);                                 break;} //BIT 6,E
         case 0x74 : {BIT(6,H);                                 break;} //BIT 6,H
         case 0x75 : {BIT(6,L);                                 break;} //BIT 6,L
         case 0x76 : {BIT(6, getMem(HL));                      break;} //BIT 6,(HL)
         case 0x77 : {BIT(6,A);                                 break;} //BIT 6,A
         case 0x78 : {BIT(7,B);                                 break;} //BIT 7,B
         case 0x79 : {BIT(7,C);                                 break;} //BIT 7,C
         case 0x7A : {BIT(7,D);                                 break;} //BIT 7,D
         case 0x7B : {BIT(7,E);                                 break;} //BIT 7,E
         case 0x7C : {BIT(7,H);                                 break;} //BIT 7,H
         case 0x7D : {BIT(7,L);                                 break;} //BIT 7,L
         case 0x7E : {BIT(7, getMem(HL));                      break;} //BIT 7,(HL)
         case 0x7F : {BIT(7,A);                                 break;} //BIT 7,A

         case 0x80 : {B=RES(0,B); BC=combine(B,C);              break;} //RES 0,B
         case 0x81 : {C=RES(0,C); BC=combine(B,C);              break;} //RES 0,C
         case 0x82 : {D=RES(0,D); DE=combine(D,E);              break;} //RES 0,D
         case 0x83 : {E=RES(0,E); DE=combine(D,E);              break;} //RES 0,E
         case 0x84 : {H=RES(0,H); HL=combine(H,L);              break;} //RES 0,H
         case 0x85 : {L=RES(0,L); HL=combine(H,L);              break;} //RES 0,L
         case 0x86 : {setMem(HL, RES(0, getMem(HL)));          break;} //RES 0,(HL)
         case 0x87 : {A=RES(0,A); AF=combine(A,F);              break;} //RES 0,A
         case 0x88 : {B=RES(1,B); BC=combine(B,C);              break;} //RES 1,B
         case 0x89 : {C=RES(1,C); BC=combine(B,C);              break;} //RES 1,C
         case 0x8A : {D=RES(1,D); DE=combine(D,E);              break;} //RES 1,D
         case 0x8B : {E=RES(1,E); DE=combine(D,E);              break;} //RES 1,E
         case 0x8C : {H=RES(1,H); HL=combine(H,L);              break;} //RES 1,H
         case 0x8D : {L=RES(1,L); HL=combine(H,L);              break;} //RES 1,L
         case 0x8E : {setMem(HL, RES(1, getMem(HL)));          break;} //RES 1,(HL)
         case 0x8F : {A=RES(1,A); AF=combine(A,F);              break;} //RES 1,A

         case 0x90 : {B=RES(2,B); BC=combine(B,C);              break;} //RES 2,B
         case 0x91 : {C=RES(2,C); BC=combine(B,C);              break;} //RES 2,C
         case 0x92 : {D=RES(2,D); DE=combine(D,E);              break;} //RES 2,D
         case 0x93 : {E=RES(2,E); DE=combine(D,E);              break;} //RES 2,E
         case 0x94 : {H=RES(2,H); HL=combine(H,L);              break;} //RES 2,H
         case 0x95 : {L=RES(2,L); HL=combine(H,L);              break;} //RES 2,L
         case 0x96 : {setMem(HL, RES(2, getMem(HL)));          break;} //RES 2,(HL)
         case 0x97 : {A=RES(2,A); AF=combine(A,F);              break;} //RES 2,A
         case 0x98 : {B=RES(3,B); BC=combine(B,C);              break;} //RES 3,B
         case 0x99 : {C=RES(3,C); BC=combine(B,C);              break;} //RES 3,C
         case 0x9A : {D=RES(3,D); DE=combine(D,E);              break;} //RES 3,D
         case 0x9B : {E=RES(3,E); DE=combine(D,E);              break;} //RES 3,E
         case 0x9C : {H=RES(3,H); HL=combine(H,L);              break;} //RES 3,H
         case 0x9D : {L=RES(3,L); HL=combine(H,L);              break;} //RES 3,L
         case 0x9E : {setMem(HL, RES(3, getMem(HL)));          break;} //RES 3,(HL)
         case 0x9F : {A=RES(3,A); AF=combine(A,F);              break;} //RES 3,A

         case 0xA0 : {B=RES(4,B); BC=combine(B,C);              break;} //RES 4,B
         case 0xA1 : {C=RES(4,C); BC=combine(B,C);              break;} //RES 4,C
         case 0xA2 : {D=RES(4,D); DE=combine(D,E);              break;} //RES 4,D
         case 0xA3 : {E=RES(4,E); DE=combine(D,E);              break;} //RES 4,E
         case 0xA4 : {H=RES(4,H); HL=combine(H,L);              break;} //RES 4,H
         case 0xA5 : {L=RES(4,L); HL=combine(H,L);              break;} //RES 4,L
         case 0xA6 : {setMem(HL, RES(4, getMem(HL)));          break;} //RES 4,(HL)
         case 0xA7 : {A=RES(4,A); AF=combine(A,F);              break;} //RES 4,A
         case 0xA8 : {B=RES(5,B); BC=combine(B,C);              break;} //RES 5,B
         case 0xA9 : {C=RES(5,C); BC=combine(B,C);              break;} //RES 5,C
         case 0xAA : {D=RES(5,D); DE=combine(D,E);              break;} //RES 5,D
         case 0xAB : {E=RES(5,E); DE=combine(D,E);              break;} //RES 5,E
         case 0xAC : {H=RES(5,H); HL=combine(H,L);              break;} //RES 5,H
         case 0xAD : {L=RES(5,L); HL=combine(H,L);              break;} //RES 5,L
         case 0xAE : {setMem(HL, RES(5, getMem(HL)));          break;} //RES 5,(HL)
         case 0xAF : {A=RES(5,A); AF=combine(A,F);              break;} //RES 5,A

         case 0xB0 : {B=RES(6,B); BC=combine(B,C);              break;} //RES 6,B
         case 0xB1 : {C=RES(6,C); BC=combine(B,C);              break;} //RES 6,C
         case 0xB2 : {D=RES(6,D); DE=combine(D,E);              break;} //RES 6,D
         case 0xB3 : {E=RES(6,E); DE=combine(D,E);              break;} //RES 6,E
         case 0xB4 : {H=RES(6,H); HL=combine(H,L);              break;} //RES 6,H
         case 0xB5 : {L=RES(6,L); HL=combine(H,L);              break;} //RES 6,L
         case 0xB6 : {setMem(HL, RES(6, getMem(HL)));          break;} //RES 6,(HL)
         case 0xB7 : {A=RES(6,A); AF=combine(A,F);              break;} //RES 6,A
         case 0xB8 : {B=RES(7,B); BC=combine(B,C);              break;} //RES 7,B
         case 0xB9 : {C=RES(7,C); BC=combine(B,C);              break;} //RES 7,C
         case 0xBA : {D=RES(7,D); DE=combine(D,E);              break;} //RES 7,D
         case 0xBB : {E=RES(7,E); DE=combine(D,E);              break;} //RES 7,E
         case 0xBC : {H=RES(7,H); HL=combine(H,L);              break;} //RES 7,H
         case 0xBD : {L=RES(7,L); HL=combine(H,L);              break;} //RES 7,L
         case 0xBE : {setMem(HL, RES(7, getMem(HL)));          break;} //RES 7,(HL)
         case 0xBF : {A=RES(7,A); AF=combine(A,F);              break;} //RES 7,A

         case 0xC0 : {B=SET(0,B); BC=combine(B,C);              break;} //SET 0,B
         case 0xC1 : {C=SET(0,C); BC=combine(B,C);              break;} //SET 0,C
         case 0xC2 : {D=SET(0,D); DE=combine(D,E);              break;} //SET 0,D
         case 0xC3 : {E=SET(0,E); DE=combine(D,E);              break;} //SET 0,E
         case 0xC4 : {H=SET(0,H); HL=combine(H,L);              break;} //SET 0,H
         case 0xC5 : {L=SET(0,L); HL=combine(H,L);              break;} //SET 0,L
         case 0xC6 : {setMem(HL, SET(0, getMem(HL)));          break;} //SET 0,(HL)
         case 0xC7 : {A=SET(0,A); AF=combine(A,F);              break;} //SET 0,A
         case 0xC8 : {B=SET(1,B); BC=combine(B,C);              break;} //SET 1,B
         case 0xC9 : {C=SET(1,C); BC=combine(B,C);              break;} //SET 1,C
         case 0xCA : {D=SET(1,D); DE=combine(D,E);              break;} //SET 1,D
         case 0xCB : {E=SET(1,E); DE=combine(D,E);              break;} //SET 1,E
         case 0xCC : {H=SET(1,H); HL=combine(H,L);              break;} //SET 1,H
         case 0xCD : {L=SET(1,L); HL=combine(H,L);              break;} //SET 1,L
         case 0xCE : {setMem(HL, SET(1, getMem(HL)));          break;} //SET 1,(HL)
         case 0xCF : {A=SET(1,A); AF=combine(A,F);              break;} //SET 1,A

         case 0xD0 : {B=SET(2,B); BC=combine(B,C);              break;} //SET 2,B
         case 0xD1 : {C=SET(2,C); BC=combine(B,C);              break;} //SET 2,C
         case 0xD2 : {D=SET(2,D); DE=combine(D,E);              break;} //SET 2,D
         case 0xD3 : {E=SET(2,E); DE=combine(D,E);              break;} //SET 2,E
         case 0xD4 : {H=SET(2,H); HL=combine(H,L);              break;} //SET 2,H
         case 0xD5 : {L=SET(2,L); HL=combine(H,L);              break;} //SET 2,L
         case 0xD6 : {setMem(HL, SET(2, getMem(HL)));          break;} //SET 2,(HL)
         case 0xD7 : {A=SET(2,A); AF=combine(A,F);              break;} //SET 2,A
         case 0xD8 : {B=SET(3,B); BC=combine(B,C);              break;} //SET 3,B
         case 0xD9 : {C=SET(3,C); BC=combine(B,C);              break;} //SET 3,C
         case 0xDA : {D=SET(3,D); DE=combine(D,E);              break;} //SET 3,D
         case 0xDB : {E=SET(3,E); DE=combine(D,E);              break;} //SET 3,E
         case 0xDC : {H=SET(3,H); HL=combine(H,L);              break;} //SET 3,H
         case 0xDD : {L=SET(3,L); HL=combine(H,L);              break;} //SET 3,L
         case 0xDE : {setMem(HL, SET(3, getMem(HL)));          break;} //SET 3,(HL)
         case 0xDF : {A=SET(3,A); AF=combine(A,F);              break;} //SET 3,A

         case 0xE0 : {B=SET(4,B); BC=combine(B,C);              break;} //SET 4,B
         case 0xE1 : {C=SET(4,C); BC=combine(B,C);              break;} //SET 4,C
         case 0xE2 : {D=SET(4,D); DE=combine(D,E);              break;} //SET 4,D
         case 0xE3 : {E=SET(4,E); DE=combine(D,E);              break;} //SET 4,E
         case 0xE4 : {H=SET(4,H); HL=combine(H,L);              break;} //SET 4,H
         case 0xE5 : {L=SET(4,L); HL=combine(H,L);              break;} //SET 4,L
         case 0xE6 : {setMem(HL, SET(4, getMem(HL)));          break;} //SET 4,(HL)
         case 0xE7 : {A=SET(4,A); AF=combine(A,F);              break;} //SET 4,A
         case 0xE8 : {B=SET(5,B); BC=combine(B,C);              break;} //SET 5,B
         case 0xE9 : {C=SET(5,C); BC=combine(B,C);              break;} //SET 5,C
         case 0xEA : {D=SET(5,D); DE=combine(D,E);              break;} //SET 5,D
         case 0xEB : {E=SET(5,E); DE=combine(D,E);              break;} //SET 5,E
         case 0xEC : {H=SET(5,H); HL=combine(H,L);              break;} //SET 5,H
         case 0xED : {L=SET(5,L); HL=combine(H,L);              break;} //SET 5,L
         case 0xEE : {setMem(HL, SET(5, getMem(HL)));          break;} //SET 5,(HL)
         case 0xEF : {A=SET(5,A); AF=combine(A,F);              break;} //SET 5,A

         case 0xF0 : {B=SET(6,B); BC=combine(B,C);              break;} //SET 6,B
         case 0xF1 : {C=SET(6,C); BC=combine(B,C);              break;} //SET 6,C
         case 0xF2 : {D=SET(6,D); DE=combine(D,E);              break;} //SET 6,D
         case 0xF3 : {E=SET(6,E); DE=combine(D,E);              break;} //SET 6,E
         case 0xF4 : {H=SET(6,H); HL=combine(H,L);              break;} //SET 6,H
         case 0xF5 : {L=SET(6,L); HL=combine(H,L);              break;} //SET 6,L
         case 0xF6 : {setMem(HL, SET(6, getMem(HL)));           break;} //SET 6,(HL)
         case 0xF7 : {A=SET(6,A); AF=combine(A,F);              break;} //SET 6,A
         case 0xF8 : {B=SET(7,B); BC=combine(B,C);              break;} //SET 7,B
         case 0xF9 : {C=SET(7,C); BC=combine(B,C);              break;} //SET 7,C
         case 0xFA : {D=SET(7,D); DE=combine(D,E);              break;} //SET 7,D
         case 0xFB : {E=SET(7,E); DE=combine(D,E);              break;} //SET 7,E
         case 0xFC : {H=SET(7,H); HL=combine(H,L);              break;} //SET 7,H
         case 0xFD : {L=SET(7,L); HL=combine(H,L);              break;} //SET 7,L
         case 0xFE : {setMem(HL, SET(7, getMem(HL)));          break;} //SET 7,(HL)
         case 0xFF : {A=SET(7,A); AF=combine(A,F);              break;} //SET 7,A

         default : {System.out.println("Unknown CB op code " + Integer.toHexString(opCode2)); break;}
      }
   }

   public boolean getThreadState() {return APPLET_RUNNING;}

   private void createDAATable()
   {
      DAA_TABLE=new int[]
      {
         0x0044,0x0100,0x0200,0x0304,0x0400,0x0504,0x0604,0x0700,
         0x0808,0x090C,0x1010,0x1114,0x1214,0x1310,0x1414,0x1510,
         0x1000,0x1104,0x1204,0x1300,0x1404,0x1500,0x1600,0x1704,
         0x180C,0x1908,0x2030,0x2134,0x2234,0x2330,0x2434,0x2530,
         0x2020,0x2124,0x2224,0x2320,0x2424,0x2520,0x2620,0x2724,
         0x282C,0x2928,0x3034,0x3130,0x3230,0x3334,0x3430,0x3534,
         0x3024,0x3120,0x3220,0x3324,0x3420,0x3524,0x3624,0x3720,
         0x3828,0x392C,0x4010,0x4114,0x4214,0x4310,0x4414,0x4510,
         0x4000,0x4104,0x4204,0x4300,0x4404,0x4500,0x4600,0x4704,
         0x480C,0x4908,0x5014,0x5110,0x5210,0x5314,0x5410,0x5514,
         0x5004,0x5100,0x5200,0x5304,0x5400,0x5504,0x5604,0x5700,
         0x5808,0x590C,0x6034,0x6130,0x6230,0x6334,0x6430,0x6534,
         0x6024,0x6120,0x6220,0x6324,0x6420,0x6524,0x6624,0x6720,
         0x6828,0x692C,0x7030,0x7134,0x7234,0x7330,0x7434,0x7530,
         0x7020,0x7124,0x7224,0x7320,0x7424,0x7520,0x7620,0x7724,
         0x782C,0x7928,0x8090,0x8194,0x8294,0x8390,0x8494,0x8590,
         0x8080,0x8184,0x8284,0x8380,0x8484,0x8580,0x8680,0x8784,
         0x888C,0x8988,0x9094,0x9190,0x9290,0x9394,0x9490,0x9594,
         0x9084,0x9180,0x9280,0x9384,0x9480,0x9584,0x9684,0x9780,
         0x9888,0x998C,0x0055,0x0111,0x0211,0x0315,0x0411,0x0515,
         0x0045,0x0101,0x0201,0x0305,0x0401,0x0505,0x0605,0x0701,
         0x0809,0x090D,0x1011,0x1115,0x1215,0x1311,0x1415,0x1511,
         0x1001,0x1105,0x1205,0x1301,0x1405,0x1501,0x1601,0x1705,
         0x180D,0x1909,0x2031,0x2135,0x2235,0x2331,0x2435,0x2531,
         0x2021,0x2125,0x2225,0x2321,0x2425,0x2521,0x2621,0x2725,
         0x282D,0x2929,0x3035,0x3131,0x3231,0x3335,0x3431,0x3535,
         0x3025,0x3121,0x3221,0x3325,0x3421,0x3525,0x3625,0x3721,
         0x3829,0x392D,0x4011,0x4115,0x4215,0x4311,0x4415,0x4511,
         0x4001,0x4105,0x4205,0x4301,0x4405,0x4501,0x4601,0x4705,
         0x480D,0x4909,0x5015,0x5111,0x5211,0x5315,0x5411,0x5515,
         0x5005,0x5101,0x5201,0x5305,0x5401,0x5505,0x5605,0x5701,
         0x5809,0x590D,0x6035,0x6131,0x6231,0x6335,0x6431,0x6535,
         0x6025,0x6121,0x6221,0x6325,0x6421,0x6525,0x6625,0x6721,
         0x6829,0x692D,0x7031,0x7135,0x7235,0x7331,0x7435,0x7531,
         0x7021,0x7125,0x7225,0x7321,0x7425,0x7521,0x7621,0x7725,
         0x782D,0x7929,0x8091,0x8195,0x8295,0x8391,0x8495,0x8591,
         0x8081,0x8185,0x8285,0x8381,0x8485,0x8581,0x8681,0x8785,
         0x888D,0x8989,0x9095,0x9191,0x9291,0x9395,0x9491,0x9595,
         0x9085,0x9181,0x9281,0x9385,0x9481,0x9585,0x9685,0x9781,
         0x9889,0x998D,0xA0B5,0xA1B1,0xA2B1,0xA3B5,0xA4B1,0xA5B5,
         0xA0A5,0xA1A1,0xA2A1,0xA3A5,0xA4A1,0xA5A5,0xA6A5,0xA7A1,
         0xA8A9,0xA9AD,0xB0B1,0xB1B5,0xB2B5,0xB3B1,0xB4B5,0xB5B1,
         0xB0A1,0xB1A5,0xB2A5,0xB3A1,0xB4A5,0xB5A1,0xB6A1,0xB7A5,
         0xB8AD,0xB9A9,0xC095,0xC191,0xC291,0xC395,0xC491,0xC595,
         0xC085,0xC181,0xC281,0xC385,0xC481,0xC585,0xC685,0xC781,
         0xC889,0xC98D,0xD091,0xD195,0xD295,0xD391,0xD495,0xD591,
         0xD081,0xD185,0xD285,0xD381,0xD485,0xD581,0xD681,0xD785,
         0xD88D,0xD989,0xE0B1,0xE1B5,0xE2B5,0xE3B1,0xE4B5,0xE5B1,
         0xE0A1,0xE1A5,0xE2A5,0xE3A1,0xE4A5,0xE5A1,0xE6A1,0xE7A5,
         0xE8AD,0xE9A9,0xF0B5,0xF1B1,0xF2B1,0xF3B5,0xF4B1,0xF5B5,
         0xF0A5,0xF1A1,0xF2A1,0xF3A5,0xF4A1,0xF5A5,0xF6A5,0xF7A1,
         0xF8A9,0xF9AD,0x0055,0x0111,0x0211,0x0315,0x0411,0x0515,
         0x0045,0x0101,0x0201,0x0305,0x0401,0x0505,0x0605,0x0701,
         0x0809,0x090D,0x1011,0x1115,0x1215,0x1311,0x1415,0x1511,
         0x1001,0x1105,0x1205,0x1301,0x1405,0x1501,0x1601,0x1705,
         0x180D,0x1909,0x2031,0x2135,0x2235,0x2331,0x2435,0x2531,
         0x2021,0x2125,0x2225,0x2321,0x2425,0x2521,0x2621,0x2725,
         0x282D,0x2929,0x3035,0x3131,0x3231,0x3335,0x3431,0x3535,
         0x3025,0x3121,0x3221,0x3325,0x3421,0x3525,0x3625,0x3721,
         0x3829,0x392D,0x4011,0x4115,0x4215,0x4311,0x4415,0x4511,
         0x4001,0x4105,0x4205,0x4301,0x4405,0x4501,0x4601,0x4705,
         0x480D,0x4909,0x5015,0x5111,0x5211,0x5315,0x5411,0x5515,
         0x5005,0x5101,0x5201,0x5305,0x5401,0x5505,0x5605,0x5701,
         0x5809,0x590D,0x6035,0x6131,0x6231,0x6335,0x6431,0x6535,
         0x0604,0x0700,0x0808,0x090C,0x0A0C,0x0B08,0x0C0C,0x0D08,
         0x0E08,0x0F0C,0x1010,0x1114,0x1214,0x1310,0x1414,0x1510,
         0x1600,0x1704,0x180C,0x1908,0x1A08,0x1B0C,0x1C08,0x1D0C,
         0x1E0C,0x1F08,0x2030,0x2134,0x2234,0x2330,0x2434,0x2530,
         0x2620,0x2724,0x282C,0x2928,0x2A28,0x2B2C,0x2C28,0x2D2C,
         0x2E2C,0x2F28,0x3034,0x3130,0x3230,0x3334,0x3430,0x3534,
         0x3624,0x3720,0x3828,0x392C,0x3A2C,0x3B28,0x3C2C,0x3D28,
         0x3E28,0x3F2C,0x4010,0x4114,0x4214,0x4310,0x4414,0x4510,
         0x4600,0x4704,0x480C,0x4908,0x4A08,0x4B0C,0x4C08,0x4D0C,
         0x4E0C,0x4F08,0x5014,0x5110,0x5210,0x5314,0x5410,0x5514,
         0x5604,0x5700,0x5808,0x590C,0x5A0C,0x5B08,0x5C0C,0x5D08,
         0x5E08,0x5F0C,0x6034,0x6130,0x6230,0x6334,0x6430,0x6534,
         0x6624,0x6720,0x6828,0x692C,0x6A2C,0x6B28,0x6C2C,0x6D28,
         0x6E28,0x6F2C,0x7030,0x7134,0x7234,0x7330,0x7434,0x7530,
         0x7620,0x7724,0x782C,0x7928,0x7A28,0x7B2C,0x7C28,0x7D2C,
         0x7E2C,0x7F28,0x8090,0x8194,0x8294,0x8390,0x8494,0x8590,
         0x8680,0x8784,0x888C,0x8988,0x8A88,0x8B8C,0x8C88,0x8D8C,
         0x8E8C,0x8F88,0x9094,0x9190,0x9290,0x9394,0x9490,0x9594,
         0x9684,0x9780,0x9888,0x998C,0x9A8C,0x9B88,0x9C8C,0x9D88,
         0x9E88,0x9F8C,0x0055,0x0111,0x0211,0x0315,0x0411,0x0515,
         0x0605,0x0701,0x0809,0x090D,0x0A0D,0x0B09,0x0C0D,0x0D09,
         0x0E09,0x0F0D,0x1011,0x1115,0x1215,0x1311,0x1415,0x1511,
         0x1601,0x1705,0x180D,0x1909,0x1A09,0x1B0D,0x1C09,0x1D0D,
         0x1E0D,0x1F09,0x2031,0x2135,0x2235,0x2331,0x2435,0x2531,
         0x2621,0x2725,0x282D,0x2929,0x2A29,0x2B2D,0x2C29,0x2D2D,
         0x2E2D,0x2F29,0x3035,0x3131,0x3231,0x3335,0x3431,0x3535,
         0x3625,0x3721,0x3829,0x392D,0x3A2D,0x3B29,0x3C2D,0x3D29,
         0x3E29,0x3F2D,0x4011,0x4115,0x4215,0x4311,0x4415,0x4511,
         0x4601,0x4705,0x480D,0x4909,0x4A09,0x4B0D,0x4C09,0x4D0D,
         0x4E0D,0x4F09,0x5015,0x5111,0x5211,0x5315,0x5411,0x5515,
         0x5605,0x5701,0x5809,0x590D,0x5A0D,0x5B09,0x5C0D,0x5D09,
         0x5E09,0x5F0D,0x6035,0x6131,0x6231,0x6335,0x6431,0x6535,
         0x6625,0x6721,0x6829,0x692D,0x6A2D,0x6B29,0x6C2D,0x6D29,
         0x6E29,0x6F2D,0x7031,0x7135,0x7235,0x7331,0x7435,0x7531,
         0x7621,0x7725,0x782D,0x7929,0x7A29,0x7B2D,0x7C29,0x7D2D,
         0x7E2D,0x7F29,0x8091,0x8195,0x8295,0x8391,0x8495,0x8591,
         0x8681,0x8785,0x888D,0x8989,0x8A89,0x8B8D,0x8C89,0x8D8D,
         0x8E8D,0x8F89,0x9095,0x9191,0x9291,0x9395,0x9491,0x9595,
         0x9685,0x9781,0x9889,0x998D,0x9A8D,0x9B89,0x9C8D,0x9D89,
         0x9E89,0x9F8D,0xA0B5,0xA1B1,0xA2B1,0xA3B5,0xA4B1,0xA5B5,
         0xA6A5,0xA7A1,0xA8A9,0xA9AD,0xAAAD,0xABA9,0xACAD,0xADA9,
         0xAEA9,0xAFAD,0xB0B1,0xB1B5,0xB2B5,0xB3B1,0xB4B5,0xB5B1,
         0xB6A1,0xB7A5,0xB8AD,0xB9A9,0xBAA9,0xBBAD,0xBCA9,0xBDAD,
         0xBEAD,0xBFA9,0xC095,0xC191,0xC291,0xC395,0xC491,0xC595,
         0xC685,0xC781,0xC889,0xC98D,0xCA8D,0xCB89,0xCC8D,0xCD89,
         0xCE89,0xCF8D,0xD091,0xD195,0xD295,0xD391,0xD495,0xD591,
         0xD681,0xD785,0xD88D,0xD989,0xDA89,0xDB8D,0xDC89,0xDD8D,
         0xDE8D,0xDF89,0xE0B1,0xE1B5,0xE2B5,0xE3B1,0xE4B5,0xE5B1,
         0xE6A1,0xE7A5,0xE8AD,0xE9A9,0xEAA9,0xEBAD,0xECA9,0xEDAD,
         0xEEAD,0xEFA9,0xF0B5,0xF1B1,0xF2B1,0xF3B5,0xF4B1,0xF5B5,
         0xF6A5,0xF7A1,0xF8A9,0xF9AD,0xFAAD,0xFBA9,0xFCAD,0xFDA9,
         0xFEA9,0xFFAD,0x0055,0x0111,0x0211,0x0315,0x0411,0x0515,
         0x0605,0x0701,0x0809,0x090D,0x0A0D,0x0B09,0x0C0D,0x0D09,
         0x0E09,0x0F0D,0x1011,0x1115,0x1215,0x1311,0x1415,0x1511,
         0x1601,0x1705,0x180D,0x1909,0x1A09,0x1B0D,0x1C09,0x1D0D,
         0x1E0D,0x1F09,0x2031,0x2135,0x2235,0x2331,0x2435,0x2531,
         0x2621,0x2725,0x282D,0x2929,0x2A29,0x2B2D,0x2C29,0x2D2D,
         0x2E2D,0x2F29,0x3035,0x3131,0x3231,0x3335,0x3431,0x3535,
         0x3625,0x3721,0x3829,0x392D,0x3A2D,0x3B29,0x3C2D,0x3D29,
         0x3E29,0x3F2D,0x4011,0x4115,0x4215,0x4311,0x4415,0x4511,
         0x4601,0x4705,0x480D,0x4909,0x4A09,0x4B0D,0x4C09,0x4D0D,
         0x4E0D,0x4F09,0x5015,0x5111,0x5211,0x5315,0x5411,0x5515,
         0x5605,0x5701,0x5809,0x590D,0x5A0D,0x5B09,0x5C0D,0x5D09,
         0x5E09,0x5F0D,0x6035,0x6131,0x6231,0x6335,0x6431,0x6535,
         0x0046,0x0102,0x0202,0x0306,0x0402,0x0506,0x0606,0x0702,
         0x080A,0x090E,0x0402,0x0506,0x0606,0x0702,0x080A,0x090E,
         0x1002,0x1106,0x1206,0x1302,0x1406,0x1502,0x1602,0x1706,
         0x180E,0x190A,0x1406,0x1502,0x1602,0x1706,0x180E,0x190A,
         0x2022,0x2126,0x2226,0x2322,0x2426,0x2522,0x2622,0x2726,
         0x282E,0x292A,0x2426,0x2522,0x2622,0x2726,0x282E,0x292A,
         0x3026,0x3122,0x3222,0x3326,0x3422,0x3526,0x3626,0x3722,
         0x382A,0x392E,0x3422,0x3526,0x3626,0x3722,0x382A,0x392E,
         0x4002,0x4106,0x4206,0x4302,0x4406,0x4502,0x4602,0x4706,
         0x480E,0x490A,0x4406,0x4502,0x4602,0x4706,0x480E,0x490A,
         0x5006,0x5102,0x5202,0x5306,0x5402,0x5506,0x5606,0x5702,
         0x580A,0x590E,0x5402,0x5506,0x5606,0x5702,0x580A,0x590E,
         0x6026,0x6122,0x6222,0x6326,0x6422,0x6526,0x6626,0x6722,
         0x682A,0x692E,0x6422,0x6526,0x6626,0x6722,0x682A,0x692E,
         0x7022,0x7126,0x7226,0x7322,0x7426,0x7522,0x7622,0x7726,
         0x782E,0x792A,0x7426,0x7522,0x7622,0x7726,0x782E,0x792A,
         0x8082,0x8186,0x8286,0x8382,0x8486,0x8582,0x8682,0x8786,
         0x888E,0x898A,0x8486,0x8582,0x8682,0x8786,0x888E,0x898A,
         0x9086,0x9182,0x9282,0x9386,0x9482,0x9586,0x9686,0x9782,
         0x988A,0x998E,0x3423,0x3527,0x3627,0x3723,0x382B,0x392F,
         0x4003,0x4107,0x4207,0x4303,0x4407,0x4503,0x4603,0x4707,
         0x480F,0x490B,0x4407,0x4503,0x4603,0x4707,0x480F,0x490B,
         0x5007,0x5103,0x5203,0x5307,0x5403,0x5507,0x5607,0x5703,
         0x580B,0x590F,0x5403,0x5507,0x5607,0x5703,0x580B,0x590F,
         0x6027,0x6123,0x6223,0x6327,0x6423,0x6527,0x6627,0x6723,
         0x682B,0x692F,0x6423,0x6527,0x6627,0x6723,0x682B,0x692F,
         0x7023,0x7127,0x7227,0x7323,0x7427,0x7523,0x7623,0x7727,
         0x782F,0x792B,0x7427,0x7523,0x7623,0x7727,0x782F,0x792B,
         0x8083,0x8187,0x8287,0x8383,0x8487,0x8583,0x8683,0x8787,
         0x888F,0x898B,0x8487,0x8583,0x8683,0x8787,0x888F,0x898B,
         0x9087,0x9183,0x9283,0x9387,0x9483,0x9587,0x9687,0x9783,
         0x988B,0x998F,0x9483,0x9587,0x9687,0x9783,0x988B,0x998F,
         0xA0A7,0xA1A3,0xA2A3,0xA3A7,0xA4A3,0xA5A7,0xA6A7,0xA7A3,
         0xA8AB,0xA9AF,0xA4A3,0xA5A7,0xA6A7,0xA7A3,0xA8AB,0xA9AF,
         0xB0A3,0xB1A7,0xB2A7,0xB3A3,0xB4A7,0xB5A3,0xB6A3,0xB7A7,
         0xB8AF,0xB9AB,0xB4A7,0xB5A3,0xB6A3,0xB7A7,0xB8AF,0xB9AB,
         0xC087,0xC183,0xC283,0xC387,0xC483,0xC587,0xC687,0xC783,
         0xC88B,0xC98F,0xC483,0xC587,0xC687,0xC783,0xC88B,0xC98F,
         0xD083,0xD187,0xD287,0xD383,0xD487,0xD583,0xD683,0xD787,
         0xD88F,0xD98B,0xD487,0xD583,0xD683,0xD787,0xD88F,0xD98B,
         0xE0A3,0xE1A7,0xE2A7,0xE3A3,0xE4A7,0xE5A3,0xE6A3,0xE7A7,
         0xE8AF,0xE9AB,0xE4A7,0xE5A3,0xE6A3,0xE7A7,0xE8AF,0xE9AB,
         0xF0A7,0xF1A3,0xF2A3,0xF3A7,0xF4A3,0xF5A7,0xF6A7,0xF7A3,
         0xF8AB,0xF9AF,0xF4A3,0xF5A7,0xF6A7,0xF7A3,0xF8AB,0xF9AF,
         0x0047,0x0103,0x0203,0x0307,0x0403,0x0507,0x0607,0x0703,
         0x080B,0x090F,0x0403,0x0507,0x0607,0x0703,0x080B,0x090F,
         0x1003,0x1107,0x1207,0x1303,0x1407,0x1503,0x1603,0x1707,
         0x180F,0x190B,0x1407,0x1503,0x1603,0x1707,0x180F,0x190B,
         0x2023,0x2127,0x2227,0x2323,0x2427,0x2523,0x2623,0x2727,
         0x282F,0x292B,0x2427,0x2523,0x2623,0x2727,0x282F,0x292B,
         0x3027,0x3123,0x3223,0x3327,0x3423,0x3527,0x3627,0x3723,
         0x382B,0x392F,0x3423,0x3527,0x3627,0x3723,0x382B,0x392F,
         0x4003,0x4107,0x4207,0x4303,0x4407,0x4503,0x4603,0x4707,
         0x480F,0x490B,0x4407,0x4503,0x4603,0x4707,0x480F,0x490B,
         0x5007,0x5103,0x5203,0x5307,0x5403,0x5507,0x5607,0x5703,
         0x580B,0x590F,0x5403,0x5507,0x5607,0x5703,0x580B,0x590F,
         0x6027,0x6123,0x6223,0x6327,0x6423,0x6527,0x6627,0x6723,
         0x682B,0x692F,0x6423,0x6527,0x6627,0x6723,0x682B,0x692F,
         0x7023,0x7127,0x7227,0x7323,0x7427,0x7523,0x7623,0x7727,
         0x782F,0x792B,0x7427,0x7523,0x7623,0x7727,0x782F,0x792B,
         0x8083,0x8187,0x8287,0x8383,0x8487,0x8583,0x8683,0x8787,
         0x888F,0x898B,0x8487,0x8583,0x8683,0x8787,0x888F,0x898B,
         0x9087,0x9183,0x9283,0x9387,0x9483,0x9587,0x9687,0x9783,
         0x988B,0x998F,0x9483,0x9587,0x9687,0x9783,0x988B,0x998F,
         0xFABE,0xFBBA,0xFCBE,0xFDBA,0xFEBA,0xFFBE,0x0046,0x0102,
         0x0202,0x0306,0x0402,0x0506,0x0606,0x0702,0x080A,0x090E,
         0x0A1E,0x0B1A,0x0C1E,0x0D1A,0x0E1A,0x0F1E,0x1002,0x1106,
         0x1206,0x1302,0x1406,0x1502,0x1602,0x1706,0x180E,0x190A,
         0x1A1A,0x1B1E,0x1C1A,0x1D1E,0x1E1E,0x1F1A,0x2022,0x2126,
         0x2226,0x2322,0x2426,0x2522,0x2622,0x2726,0x282E,0x292A,
         0x2A3A,0x2B3E,0x2C3A,0x2D3E,0x2E3E,0x2F3A,0x3026,0x3122,
         0x3222,0x3326,0x3422,0x3526,0x3626,0x3722,0x382A,0x392E,
         0x3A3E,0x3B3A,0x3C3E,0x3D3A,0x3E3A,0x3F3E,0x4002,0x4106,
         0x4206,0x4302,0x4406,0x4502,0x4602,0x4706,0x480E,0x490A,
         0x4A1A,0x4B1E,0x4C1A,0x4D1E,0x4E1E,0x4F1A,0x5006,0x5102,
         0x5202,0x5306,0x5402,0x5506,0x5606,0x5702,0x580A,0x590E,
         0x5A1E,0x5B1A,0x5C1E,0x5D1A,0x5E1A,0x5F1E,0x6026,0x6122,
         0x6222,0x6326,0x6422,0x6526,0x6626,0x6722,0x682A,0x692E,
         0x6A3E,0x6B3A,0x6C3E,0x6D3A,0x6E3A,0x6F3E,0x7022,0x7126,
         0x7226,0x7322,0x7426,0x7522,0x7622,0x7726,0x782E,0x792A,
         0x7A3A,0x7B3E,0x7C3A,0x7D3E,0x7E3E,0x7F3A,0x8082,0x8186,
         0x8286,0x8382,0x8486,0x8582,0x8682,0x8786,0x888E,0x898A,
         0x8A9A,0x8B9E,0x8C9A,0x8D9E,0x8E9E,0x8F9A,0x9086,0x9182,
         0x9282,0x9386,0x3423,0x3527,0x3627,0x3723,0x382B,0x392F,
         0x3A3F,0x3B3B,0x3C3F,0x3D3B,0x3E3B,0x3F3F,0x4003,0x4107,
         0x4207,0x4303,0x4407,0x4503,0x4603,0x4707,0x480F,0x490B,
         0x4A1B,0x4B1F,0x4C1B,0x4D1F,0x4E1F,0x4F1B,0x5007,0x5103,
         0x5203,0x5307,0x5403,0x5507,0x5607,0x5703,0x580B,0x590F,
         0x5A1F,0x5B1B,0x5C1F,0x5D1B,0x5E1B,0x5F1F,0x6027,0x6123,
         0x6223,0x6327,0x6423,0x6527,0x6627,0x6723,0x682B,0x692F,
         0x6A3F,0x6B3B,0x6C3F,0x6D3B,0x6E3B,0x6F3F,0x7023,0x7127,
         0x7227,0x7323,0x7427,0x7523,0x7623,0x7727,0x782F,0x792B,
         0x7A3B,0x7B3F,0x7C3B,0x7D3F,0x7E3F,0x7F3B,0x8083,0x8187,
         0x8287,0x8383,0x8487,0x8583,0x8683,0x8787,0x888F,0x898B,
         0x8A9B,0x8B9F,0x8C9B,0x8D9F,0x8E9F,0x8F9B,0x9087,0x9183,
         0x9283,0x9387,0x9483,0x9587,0x9687,0x9783,0x988B,0x998F,
         0x9A9F,0x9B9B,0x9C9F,0x9D9B,0x9E9B,0x9F9F,0xA0A7,0xA1A3,
         0xA2A3,0xA3A7,0xA4A3,0xA5A7,0xA6A7,0xA7A3,0xA8AB,0xA9AF,
         0xAABF,0xABBB,0xACBF,0xADBB,0xAEBB,0xAFBF,0xB0A3,0xB1A7,
         0xB2A7,0xB3A3,0xB4A7,0xB5A3,0xB6A3,0xB7A7,0xB8AF,0xB9AB,
         0xBABB,0xBBBF,0xBCBB,0xBDBF,0xBEBF,0xBFBB,0xC087,0xC183,
         0xC283,0xC387,0xC483,0xC587,0xC687,0xC783,0xC88B,0xC98F,
         0xCA9F,0xCB9B,0xCC9F,0xCD9B,0xCE9B,0xCF9F,0xD083,0xD187,
         0xD287,0xD383,0xD487,0xD583,0xD683,0xD787,0xD88F,0xD98B,
         0xDA9B,0xDB9F,0xDC9B,0xDD9F,0xDE9F,0xDF9B,0xE0A3,0xE1A7,
         0xE2A7,0xE3A3,0xE4A7,0xE5A3,0xE6A3,0xE7A7,0xE8AF,0xE9AB,
         0xEABB,0xEBBF,0xECBB,0xEDBF,0xEEBF,0xEFBB,0xF0A7,0xF1A3,
         0xF2A3,0xF3A7,0xF4A3,0xF5A7,0xF6A7,0xF7A3,0xF8AB,0xF9AF,
         0xFABF,0xFBBB,0xFCBF,0xFDBB,0xFEBB,0xFFBF,0x0047,0x0103,
         0x0203,0x0307,0x0403,0x0507,0x0607,0x0703,0x080B,0x090F,
         0x0A1F,0x0B1B,0x0C1F,0x0D1B,0x0E1B,0x0F1F,0x1003,0x1107,
         0x1207,0x1303,0x1407,0x1503,0x1603,0x1707,0x180F,0x190B,
         0x1A1B,0x1B1F,0x1C1B,0x1D1F,0x1E1F,0x1F1B,0x2023,0x2127,
         0x2227,0x2323,0x2427,0x2523,0x2623,0x2727,0x282F,0x292B,
         0x2A3B,0x2B3F,0x2C3B,0x2D3F,0x2E3F,0x2F3B,0x3027,0x3123,
         0x3223,0x3327,0x3423,0x3527,0x3627,0x3723,0x382B,0x392F,
         0x3A3F,0x3B3B,0x3C3F,0x3D3B,0x3E3B,0x3F3F,0x4003,0x4107,
         0x4207,0x4303,0x4407,0x4503,0x4603,0x4707,0x480F,0x490B,
         0x4A1B,0x4B1F,0x4C1B,0x4D1F,0x4E1F,0x4F1B,0x5007,0x5103,
         0x5203,0x5307,0x5403,0x5507,0x5607,0x5703,0x580B,0x590F,
         0x5A1F,0x5B1B,0x5C1F,0x5D1B,0x5E1B,0x5F1F,0x6027,0x6123,
         0x6223,0x6327,0x6423,0x6527,0x6627,0x6723,0x682B,0x692F,
         0x6A3F,0x6B3B,0x6C3F,0x6D3B,0x6E3B,0x6F3F,0x7023,0x7127,
         0x7227,0x7323,0x7427,0x7523,0x7623,0x7727,0x782F,0x792B,
         0x7A3B,0x7B3F,0x7C3B,0x7D3F,0x7E3F,0x7F3B,0x8083,0x8187,
         0x8287,0x8383,0x8487,0x8583,0x8683,0x8787,0x888F,0x898B,
         0x8A9B,0x8B9F,0x8C9B,0x8D9F,0x8E9F,0x8F9B,0x9087,0x9183,
         0x9283,0x9387,0x9483,0x9587,0x9687,0x9783,0x988B,0x998F
      };
   }

   private void createSwapTable()
   {
      SWAP_TABLE=new char[]
      {
         0x00,0x10,0x20,0x30,0x40,0x50,0x60,0x70,0x80,0x90,0xA0,0xB0,0xC0,0xD0,0xE0,0xF0,
         0x01,0x11,0x21,0x31,0x41,0x51,0x61,0x71,0x81,0x91,0xA1,0xB1,0xC1,0xD1,0xE1,0xF1,
         0x02,0x12,0x22,0x32,0x42,0x52,0x62,0x72,0x82,0x92,0xA2,0xB2,0xC2,0xD2,0xE2,0xF2,
         0x03,0x13,0x23,0x33,0x43,0x53,0x63,0x73,0x83,0x93,0xA3,0xB3,0xC3,0xD3,0xE3,0xF3,
         0x04,0x14,0x24,0x34,0x44,0x54,0x64,0x74,0x84,0x94,0xA4,0xB4,0xC4,0xD4,0xE4,0xF4,
         0x05,0x15,0x25,0x35,0x45,0x55,0x65,0x75,0x85,0x95,0xA5,0xB5,0xC5,0xD5,0xE5,0xF5,
         0x06,0x16,0x26,0x36,0x46,0x56,0x66,0x76,0x86,0x96,0xA6,0xB6,0xC6,0xD6,0xE6,0xF6,
         0x07,0x17,0x27,0x37,0x47,0x57,0x67,0x77,0x87,0x97,0xA7,0xB7,0xC7,0xD7,0xE7,0xF7,
         0x08,0x18,0x28,0x38,0x48,0x58,0x68,0x78,0x88,0x98,0xA8,0xB8,0xC8,0xD8,0xE8,0xF8,
         0x09,0x19,0x29,0x39,0x49,0x59,0x69,0x79,0x89,0x99,0xA9,0xB9,0xC9,0xD9,0xE9,0xF9,
         0x0A,0x1A,0x2A,0x3A,0x4A,0x5A,0x6A,0x7A,0x8A,0x9A,0xAA,0xBA,0xCA,0xDA,0xEA,0xFA,
         0x0B,0x1B,0x2B,0x3B,0x4B,0x5B,0x6B,0x7B,0x8B,0x9B,0xAB,0xBB,0xCB,0xDB,0xEB,0xFB,
         0x0C,0x1C,0x2C,0x3C,0x4C,0x5C,0x6C,0x7C,0x8C,0x9C,0xAC,0xBC,0xCC,0xDC,0xEC,0xFC,
         0x0D,0x1D,0x2D,0x3D,0x4D,0x5D,0x6D,0x7D,0x8D,0x9D,0xAD,0xBD,0xCD,0xDD,0xED,0xFD,
         0x0E,0x1E,0x2E,0x3E,0x4E,0x5E,0x6E,0x7E,0x8E,0x9E,0xAE,0xBE,0xCE,0xDE,0xEE,0xFE,
         0x0F,0x1F,0x2F,0x3F,0x4F,0x5F,0x6F,0x7F,0x8F,0x9F,0xAF,0xBF,0xCF,0xDF,0xEF,0xFF
      };
   }
}
